/*
 * Copyright (C) 2006-2007 by basyskom GmbH
 *  @author Mike Hauth <info@basyskom.de>
 *  @author Eva Brucherseifer <info@basyskom.de>
 *  @author Tobias Hunger <info@basyskom.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "callmodel.h"
#include "decibel_log.h"
#include "kcallcore.h"

#include <Decibel/Types>
#include <Decibel/DBusNames>

#include <QtTapioca/StreamChannel>
#include <QtTapioca/ContactGroup>
#include <QtTapioca/UserContact>
#include <QtTapioca/Connection>
#include <QtTapioca/Contact>

#include <QtDBus/QDBusConnection>
#include <QtDBus/QDBusReply>

#include <QtGlobal>

#include <kdebug.h>
#include <kstandarddirs.h>
#include <klocale.h>

#include <QString>
#include <QList>
#include <QFile>
#include <QDateTime>

#include <QPixmap>

class CallModelPrivate
{
public:
    CallModelPrivate(KCallCore * const core) :
        m_core(core)
    { Q_ASSERT(0 != m_core); }

    ~CallModelPrivate()
    { }

    struct ChannelInfo
    {
        ChannelInfo(QtTapioca::Connection * conn,
                    QtTapioca::StreamChannel * chan,
                    const CallStatus stat,
                    const QString & cnr,
                    const QString & cname,
                    const QString & sub) :
            connection(conn), channel(chan), status(stat),
            caller_number(cnr), caller_name(cname),
            subject(sub), startTime(QDateTime::currentDateTime())
        { }

        ChannelInfo(const ChannelInfo & element) :
            connection(element.connection),
            channel(element.channel),
            status(element.status),
            caller_number(element.caller_number),
            caller_name(element.caller_name),
            subject(element.subject),
            startTime(element.startTime)
        { }

        QtTapioca::Connection * connection;
        QtTapioca::StreamChannel * channel;
        CallStatus status;
        QString caller_number;
        QString caller_name;
        QString subject;
        QDateTime startTime;
    };

    int findChannelInfo(const QString & id)
    {
        qDebug() << "CallModel: Searching through" << m_channels.size() << "channels.";
        for (int i = 0; i < m_channels.size(); ++i)
        {
            qDebug() << "***" << m_channels[i].channel->serviceName() << "==" << id;
            if (m_channels[i].channel->serviceName() == id)
            { return i; }
        }
        return -1;
    }

    int findChannelInfo(const QtTapioca::StreamChannel * chan)
    {
        qDebug() << "CallModel: Searching through" << m_channels.size() << "channels.";
        for (int i = 0; i < m_channels.size(); ++i)
        {
            qDebug() << "***" << (void *)(m_channels[i].channel) << "==" << (void *)(chan);
            if (m_channels[i].channel == chan)
            { return i; }
        }
        return -1;
    }

    KCallCore * const m_core;

    QList<ChannelInfo> m_channels;
};

// ------------------------------------------------------

CallModel::CallModel(KCallCore * const core, QObject * parent) :
    QAbstractTableModel(parent),
    d(new CallModelPrivate(core))
{
    kDebug() << "CallModel::CallModel()" << endl;
}

CallModel::~CallModel()
{ delete d; }


int CallModel::rowCount(const QModelIndex & /* parent */) const
{ return d->m_channels.size(); }

int CallModel::columnCount(const QModelIndex & /* parent */) const
{ return COLUMN_MAX; }

QVariant CallModel::data(const QModelIndex & index,
                         int role) const
{
    if (!index.isValid()) { return QVariant(); }
    int row = index.row();
    int col = index.column();

    if (row >= rowCount()) { return QVariant(); }
    if (col >= columnCount()) { return QVariant(); }

    switch (col)
    {
    case CALL_ID:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].channel->serviceName()); }
        else
        { return QVariant(); }
    case CALLER_NR:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].caller_number); }
        else
        { return QVariant(); }
    case CALLER_NAME:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].caller_name); }
        else
        { return QVariant(); }
    case SUBJECT:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].subject); }
        else
        { return QVariant(); }
    case STATUS:
        if ( Qt::DecorationRole == role)
        { return QVariant(d->m_core->getStatusIcon(d->m_channels[row].status)); }
        else if (Qt::UserRole == role)
        { return QVariant(d->m_channels[row].status); }
        else
        { return QVariant(); }
    case DATE:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].startTime.date()); }
        else
        { return QVariant(); }
    case TIME:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_channels[row].startTime.time()); }
        else
        { return QVariant(); }
    default:
        kWarning() << "Missing column definition!" << endl;
    }
    return QVariant();
}

QVariant CallModel::headerData (int section, Qt::Orientation orientation,
                                   int role) const
{
    if (Qt::DisplayRole != role) { return QVariant(); }
    if (orientation == Qt::Horizontal)
    {
        switch (section)
        {
        case CALL_ID:
            return QVariant(i18n("Call ID"));
        case CALLER_NR:
            return QVariant(i18n("Number"));
        case CALLER_NAME:
            return QVariant(i18n("Name"));
        case SUBJECT:
            return QVariant(i18n("Subject"));
        case STATUS:
            return QVariant(i18n("Status"));
        case DATE:
            return QVariant(i18n("Date"));
        case TIME:
            return QVariant(i18n("Time"));
        default:
            return QVariant();
        }
    }
    else
    { return QVariant(QString::number(section)); }
}

// Custom methods:

void CallModel::addChannel(QtTapioca::Connection * const connection,
                           QtTapioca::StreamChannel * const stream_channel,
                           const bool incoming)
{
    qDebug() << "CallModel::addChannel.";
    Q_ASSERT(0 != connection);
    Q_ASSERT(0 != stream_channel);

    connect(stream_channel, SIGNAL(closed()), this, SLOT(onChannelClosed()));

    QtTapioca::ContactGroup * contact_group = stream_channel->contactGroup();
    // All stream channels must support the Group interface:
    // Yes, I know: this requirement from telepathy is not documented
    // anywhere...
    if (0 == contact_group) { return; }

    connect(contact_group, SIGNAL(contactEntered(QtTapioca::Contact *)),
            this, SLOT(onContactChanged(QtTapioca::Contact *)));
    connect(contact_group, SIGNAL(contactLeft(QtTapioca::Contact *)),
            this, SLOT(onContactChanged(QtTapioca::Contact *)));

    // Find contacts:
    QStringList contacts;
    QtTapioca::UserContact * self_contact = connection->userContact();
    QList<QtTapioca::Contact *> contact_ptrs(contact_group->contacts());
    QtTapioca::Contact * current_contact_ptr;
    foreach (current_contact_ptr, contact_ptrs)
    {
        if ((*current_contact_ptr) == (*self_contact))
        { continue; }
        contacts.append(current_contact_ptr->uri());
    }
    contact_ptrs = contact_group->pendingContacts();
    foreach (current_contact_ptr, contact_ptrs)
    {
        if ((*current_contact_ptr) == (*self_contact))
        { continue; }
        contacts.append(current_contact_ptr->uri());
    }

    if (contacts.isEmpty()) { return; }

    QString caller_name;
    if (!contacts[0].isEmpty())
    {
        KABC::Addressee person = d->m_core->getPersonFromNumber(contacts[0]);
        if(!person.isEmpty())
        { caller_name = person.formattedName(); }
    }

    beginInsertRows(QModelIndex(), d->m_channels.size(), d->m_channels.size());
    d->m_channels.append(CallModelPrivate::ChannelInfo(
                         connection, stream_channel,
                         (incoming) ? INCOMING : OUTGOING,
                         contacts[0], caller_name, QString()));
    endInsertRows();
}

void CallModel::onChannelClosed()
{
    QtTapioca::StreamChannel * stream_channel = dynamic_cast<QtTapioca::StreamChannel *>(sender());
    if (0 == stream_channel) { return; }

    // todo: update views.
    int pos;
    for (pos = 0; pos < d->m_channels.size(); ++pos)
    {
        if (d->m_channels[pos].channel == stream_channel)
        { break; }
    }
    if (pos < d->m_channels.size())
    {
        beginRemoveRows(QModelIndex(), pos, pos);
        d->m_channels.removeAt(pos);
        endRemoveRows();
    }
}

void CallModel::onContactChanged(QtTapioca::Contact * /* contact */)
{
    qDebug() << "CallModel::onContactChanged.";
    QtTapioca::ContactGroup * contact_group =
            dynamic_cast<QtTapioca::ContactGroup *>(sender());
    if (contact_group == 0)
    {
        qDebug() << "CallModel::onContactChanged: Failed to get sender group.";
        return;
    }
    int pos = d->findChannelInfo(dynamic_cast<QtTapioca::StreamChannel *>(contact_group->channel()));
    if (-1 == pos)
    {
        qDebug() << "CallMode::onContactChanged: ChannelInfo not found!";
        return;
    }

    CallModelPrivate::ChannelInfo * info(&(d->m_channels[pos]));
    if (0 == info->channel) { return; }

    switch (info->status)
    {
    case INCOMING:
    case OUTGOING:
        if (info->channel->contactGroup()->contacts().size() > 1)
        {
            info->status = ACTIVE;
            qDebug() << "CallMode::onContactChanged: getting ACTIVE!";
            QModelIndex idx(index(pos, int(STATUS)));
            emit dataChanged(idx, idx);
        }
        break;
    case ACTIVE:
        if (info->channel->contactGroup()->contacts().size() < 2)
        {
            info->status = HANGUP;
            qDebug() << "CallMode::onContactChanged: HANGING UP!";
            QModelIndex idx(index(pos, int(STATUS)));
            emit dataChanged(idx, idx);
        }
        break;
    default:
        qDebug() << "CallMode::onContactChanged: doing nothing!";
        break;
    }
}

void CallModel::acceptCall(const QString & id)
{
    qDebug() << "CallModel::acceptCall";
    int pos = d->findChannelInfo(id);
    if (-1 == pos) { return; }

    CallModelPrivate::ChannelInfo * info(&(d->m_channels[pos]));
    Q_ASSERT(0 != info->channel);

    switch (info->status)
    {
    case INCOMING:
    case OUTGOING:
        {
            QList<QtTapioca::Contact *> pending_contacts(info->channel->contactGroup()->pendingContacts());
            QtTapioca::Contact * contact;
            foreach (contact, pending_contacts)
            { info->channel->contactGroup()->inviteContact(contact); }
        }
        break;
    default:
        break;
    }
}

void CallModel::hangupCall(const QString & id)
{
    qDebug() << "CallModel::hangupCall";
    int pos = d->findChannelInfo(id);
    if (-1 == pos) { return; }

    CallModelPrivate::ChannelInfo * info(&(d->m_channels[pos]));
    Q_ASSERT(0 != info->channel);

    QModelIndex idx(index(pos, int(STATUS)));
    switch (info->status)
    {
    case ACTIVE:
    case INCOMING:
    case OUTGOING:
        info->channel->close();
        info->status = HANGUP;
        emit dataChanged(idx, idx);
        break;
    default:
        break;
    }
}

bool CallModel::hasDTMFSupport(const QString & id)
{
    qDebug() << "CallModel::hasDTMFSupport";
    int pos = d->findChannelInfo(id);
    if (-1 == pos) { return false; }

    CallModelPrivate::ChannelInfo * info(&(d->m_channels[pos]));
    Q_ASSERT(0 != info->channel);

    return info->channel->hasDTMFSupport();
}

void CallModel::sendDTMF(const QString & id, const QChar tone)
{
    qDebug() << "CallModel::sendDTMF";
    int pos = d->findChannelInfo(id);
    if (-1 == pos) { return; }

    CallModelPrivate::ChannelInfo * info(&(d->m_channels[pos]));
    Q_ASSERT(0 != info->channel);

    QModelIndex idx(index(pos, int(STATUS)));
    switch (info->status)
    {
    case ACTIVE:
        if (info->channel->hasDTMFSupport())
        {
            uchar event;
            if (tone == '0')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_0; }
            else if (tone == '1')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_1; }
            else if (tone == '2')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_2; }
            else if (tone == '3')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_3; }
            else if (tone == '4')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_4; }
            else if (tone == '5')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_5; }
            else if (tone == '6')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_6; }
            else if (tone == '7')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_7; }
            else if (tone == '8')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_8; }
            else if (tone == '9')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Digit_9; }
            else if (tone == '*')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Asterisk; }
            else if (tone == '#')
            { event = org::freedesktop::Telepathy::DTMF_EVENT_Hash; }
            else { return; }
            info->channel->sendDTMFTone(event);
        }
        break;
    case INCOMING:
    case OUTGOING:
        break;
    default:
        break;
    }
}
