/*
 * Copyright (C) 2006-2007 by basyskom GmbH
 *  @author Mike Hauth <info@basyskom.de>
 *  @author Eva Brucherseifer <info@basyskom.de>
 *  @author Tobias Hunger <info@basyskom.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "historymodel.h"
#include "decibel_log.h"
#include "kcallcore.h"

#include <Decibel/Types>
#include <Decibel/DBusNames>

#include <QtDBus/QDBusConnection>
#include <QtDBus/QDBusReply>

#include <QtGlobal>

#include <kdebug.h>
#include <kstandarddirs.h>
#include <kiconloader.h>
#include <klocale.h>

#include <QString>
#include <QList>
#include <QFile>
#include <QDateTime>
#include <QPixmap>

#include <Decibel/Types>

class HistoryModelPrivate
{
public:
    HistoryModelPrivate(KCallCore * const core) :
        m_core(core),
        m_log(0)
    {
        Q_ASSERT(0 != m_core);

        Decibel::registerTypes();
        m_log = new de::basyskom::Decibel::Log(Decibel::log_service, Decibel::log_path,
                                               QDBusConnection::sessionBus(), 0);
        Q_ASSERT(0 != m_log);
    }

    ~HistoryModelPrivate()
    { }

    KCallCore * const m_core;
    de::basyskom::Decibel::Log * m_log;
    QList<CallInfo> m_entries;

    // Helper methods:

    /** @brief Startup initialization.
     *
     * Get all known entries from the Decibel Log service.
     */
    void initializeHistory()
    {
        QDBusReply<uint> size = m_log->numberOfEntries();
        if (!size.isValid())
        {
            QDBusError error = size.error();
            qWarning() << "DBus Error while getting call list:"
                       << error.type() << ":"
                       << error.message();
            return;
        }

        for(uint i = 0; i < size.value(); ++i)
        { addEntry(i); }
    }

    /**
     * @brief Process one log entry.
     * @param pos The position of the entry.
     *
     * Process one entry from the Decibel Log service provider and
     * add it into the internal list of calls.
     */
    void addEntry(const uint pos)
    {
        QDBusReply<Decibel::LogEntry> entry_reply = m_log->getEntry(pos);
        if (!entry_reply.isValid())
        {
            QDBusError error = entry_reply.error();
            qWarning() << "DBus Error while log data:"
                       << error.type() << ":"
                       << error.message();
            return;
        }

        Decibel::LogEntry entry(entry_reply.value());
        CallInfo info;

        info.callId = QString::number(pos);
        if (entry.contacts.isEmpty()) { info.callerNr = QString(); }
        else { info.callerNr = entry.contacts[0]; }
        info.subject = QString();
        info.status = (entry.incoming) ? INCOMING : OUTGOING;

        QDateTime startTime;
        if (entry.connectTime == 0)
        { startTime = QDateTime::fromTime_t(entry.startTime); }
        else
        { startTime = QDateTime::fromTime_t(entry.connectTime); }
        info.date = startTime.date();
        info.time = startTime.time();

        QDateTime endTime(QDateTime::fromTime_t(entry.endTime));
        info.duration = startTime.secsTo(endTime);
        if (info.duration < 0) { info.duration = 0; }

        if (!info.callerNr.isEmpty())
        {
            KABC::Addressee person = m_core->getPersonFromNumber(info.callerNr);
            if(!person.isEmpty())
            {
                info.callerName = person.formattedName();
            }
        }
        else
        { info.callerName = QString(); }

        m_entries.append(info);
    }
};

// ------------------------------------------------------

HistoryModel::HistoryModel(KCallCore * const core, QObject * parent) :
    QAbstractTableModel(parent),
    d(new HistoryModelPrivate(core))
{
    kDebug() << "HistoryModel::HistoryModel()" << endl;
    connect(d->m_log, SIGNAL(logCleared()), this, SLOT(doLogCleared()));
    connect(d->m_log, SIGNAL(newEntry(uint)),
            this, SLOT(doEntryAdded(const uint)));
    kDebug() << "HistoryModel::HistoryModel(): connected!" << endl;

    d->initializeHistory();
}

HistoryModel::~HistoryModel()
{
    kDebug() << "HistoryModel::~HistoryModel()" << endl;
    delete d;
}


int HistoryModel::rowCount(const QModelIndex & /* parent */) const
{ return d->m_entries.size(); }

int HistoryModel::columnCount(const QModelIndex & /* parent */) const
{ return COLUMN_MAX; }

QVariant HistoryModel::data(const QModelIndex & index,
                            int role) const
{
    if (!index.isValid()) { return QVariant(); }
    int row = index.row();
    int col = index.column();

    if (row >= rowCount()) { return QVariant(); }
    if (col >= columnCount()) { return QVariant(); }

    switch (col)
    {
    case CALL_ID:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].callId); }
        else
        { return QVariant(); }
    case CALLER_NR:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].callerNr); }
        else
        { return QVariant(); }
    case CALLER_NAME:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].callerName); }
        else
        { return QVariant(); }
    case SUBJECT:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].subject); }
        else
        { return QVariant(); }
    case STATUS:
        if ( Qt::DecorationRole == role)
        { return QVariant(d->m_core->getStatusIcon(d->m_entries[row].status)); }
        else
        { return QVariant(); }
    case DATE:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].date); }
        else
        { return QVariant(); }
    case TIME:
        if ( Qt::DisplayRole == role)
        { return QVariant(d->m_entries[row].time); }
        else
        { return QVariant(); }
    case DURATION:
        if ( Qt::DisplayRole == role)
        {
            QDateTime durationTime;
            int durInt(d->m_entries[row].duration);

            durationTime.setTime_t(durInt);
            QString duration = "00:00";
            if(durInt < 3600)
            duration = durationTime.toString("mm:ss");
            else if(durInt < 3600*24)
            duration = durationTime.toString("hh:mm:ss");
            else
            duration = durationTime.toString("dd:hh:mm:ss");
            return QVariant(duration);
        }
        else
        { return QVariant(); }
    default:
        kWarning() << "Missing column definition!" << endl;
    }
    return QVariant();
}

QVariant HistoryModel::headerData (int section, Qt::Orientation orientation,
                                   int role) const
{
    if (Qt::DisplayRole != role) { return QVariant(); }
    if (orientation == Qt::Horizontal)
    {
        switch (section)
        {
        case CALL_ID:
            return QVariant(i18n("Call ID"));
        case CALLER_NR:
            return QVariant(i18n("Number"));
        case CALLER_NAME:
            return QVariant(i18n("Name"));
        case SUBJECT:
            return QVariant(i18n("Subject"));
        case STATUS:
            return QVariant(i18n("Status"));
        case DATE:
            return QVariant(i18n("Date"));
        case TIME:
            return QVariant(i18n("Time"));
        case DURATION:
            return QVariant(i18n("Duration"));
        default:
            return QVariant();
        }
    }
    else
    { return QVariant(QString::number(section)); }
}

// Custom methods:

void HistoryModel::setSubject(const int row, const QString & subject)
{
    if (row >= rowCount()) { return; }
    d->m_entries[row].subject = subject;
    QModelIndex changed_index = index(row, int(SUBJECT));
    emit dataChanged(changed_index, changed_index);
}

// Slots (for D-Bus):

void HistoryModel::doLogCleared()
{
    d->m_entries.clear();
    reset();
    kDebug() << "HistoryModel::doLogCleared" << endl;
}

void HistoryModel::doEntryAdded(const uint pos)
{
    beginInsertRows(QModelIndex(), pos, pos);
    d->addEntry(pos);
    endInsertRows();
    kDebug() << "HistoryModel::doEntryAdded" << endl;
}
