/*
 * This file is part of Akregator
 *
 * Copyright (C) 2009 Frank Osterfeld <osterfeld@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "migratefeedscommand.h"
#include "migratefeedswizard_p.h"
#include "importitemsjob.h"

#include "libkrss/importopmljob.h"
#include "libkrss/resource.h"
#include "libkrss/feedlist.h"
#include "libkrss/netfeed.h"
#include "libkrss/resourcemanager.h"

#include <QComboBox>
#include <QFile>
#include <QLabel>
#include <QListView>
#include <QPointer>
#include <QString>
#include <QStringList>
#include <QStandardItemModel>
#include <QVBoxLayout>
#include <QWidget>
#include <QWizard>
#include <QWizardPage>

#include <KDebug>
#include <KLocalizedString>
#include <KIcon>
#include <KUrl>

#include <boost/shared_ptr.hpp>

#include <cassert>

using namespace boost;
using namespace Akregator;
using namespace KRss;

StartPage::StartPage( const QStringList& resources, QWidget* parent ) : QWizardPage( parent ), m_resourceBox( 0 ) {
    setTitle( i18n("Akonadi Data Migration" ) );
    QVBoxLayout* layout = new QVBoxLayout( this );
    QLabel* label = new QLabel( i18n("Since version 1.4, Akregator uses the Akonadi framework to store feeds and articles. Your feeds and articles from older Akregator versions will now be migrated." ) );
    label->setWordWrap( true );
    layout->addWidget( label );
    if ( !resources.isEmpty() ) {
        QLabel* resLabel = new QLabel( i18n("Please choose the resource to use for importing feeds and articles." ) );
        layout->addWidget( resLabel );
        QWidget* hbox = new QWidget;
        QHBoxLayout* hlayout = new QHBoxLayout( hbox );
        m_resourceBox = new QComboBox;
        Q_FOREACH( const QString& i, resources )
            m_resourceBox->addItem( i, i );
        m_resourceBox->setCurrentIndex( 0 );
        hlayout->addWidget( new QLabel( i18n("Resource:" ) ) );
        hlayout->addWidget( m_resourceBox );
        layout->addWidget( hbox );
    }
}

OpmlImportResultPage::OpmlImportResultPage( QWidget* parent ) : WizardPage( parent ) {
    QVBoxLayout* layout = new QVBoxLayout( this );
    m_progressBar = new QProgressBar;
    layout->addWidget( m_progressBar );
    m_resultLabel = new QLabel;
    m_resultLabel->setWordWrap( true );
    layout->addWidget( m_resultLabel );
}

ItemImportResultPage::ItemImportResultPage( QWidget* parent ) : WizardPage( parent ) {
    QVBoxLayout* layout = new QVBoxLayout( this );
    layout->addWidget( m_progressBar = new QProgressBar );
    layout->addWidget( m_progressLabel = new QLabel );
    m_progressLabel->setWordWrap( true );
    m_model = new QStandardItemModel( this );
    QListView* listView = new QListView;
    layout->addWidget( listView );
    listView->setModel( m_model );
}

void ItemImportResultPage::setState( ItemImportResultPage::State state ) {
    m_progressBar->setRange( 0, state == Running ? 0 : 100 );
    m_progressBar->setValue( state == Finished ? 100 : 0 );
}

void OpmlImportResultPage::setState( OpmlImportResultPage::State state ) {
    m_progressBar->setRange( 0, state == Running ? 0 : 100 );
    m_progressBar->setValue( state == Finished ? 100 : 0 );
}

void ItemImportResultPage::connectJob( Akregator::ImportItemsJob* job, const QString& title ) {
    m_title = title;
    connect( job, SIGNAL(finished(KJob*)), this, SLOT(itemImportFinished(KJob*)) );
}

void ItemImportResultPage::itemImportFinished( KJob* j ) {
    const Akregator::ImportItemsJob* const job = qobject_cast<Akregator::ImportItemsJob*>( j );
    assert( job );

    kDebug() << m_title << job->error() << job->errorText();

    QStandardItem* item = new QStandardItem;

    if ( job->error() ) {
        item->setText( i18n( "Failed: %1 (%2)", m_title, job->errorText() ) );
        item->setIcon( KIcon("dialog-error") );
    }
    else {
        item->setText( i18n( "Success: %1", m_title ) );
        item->setIcon( KIcon("dialog-ok") );
    }
    m_model->appendRow( item );
}

namespace {
    struct FeedEntry {
        QString title;
    };
}

class Akregator::MigrateFeedsCommand::Private {
    MigrateFeedsCommand* const q;
public:
    explicit Private( MigrateFeedsCommand* qq ) : q( qq ), opmlImportResultPageId( -1 ), itemImportResultPageId( -1 ) {}
    ~Private() {
        delete wizard;
    }
    void doStart();
    void currentIdChanged( int id );
    void opmlImportFinished( KJob* job );

    void ensureWizardCreated();
    void ensureWizardShown();

    void determineResource();
    void startOpmlImport();
    void startNextItemImport();
    void itemImportFinished( KJob* );
    void resourceCreated( const QString& id );
    void wizardClosed();
    void feedAdded( Feed* );
    void feedListRetrievalFinished( KJob* );

    QPointer<QWidget> parentWidget;
    QPointer<QWizard> wizard;
    QPointer<OpmlImportResultPage> opmlImportResultPage;
    int opmlImportResultPageId;
    QPointer<ItemImportResultPage> itemImportResultPage;
    int itemImportResultPageId;
    QPointer<StartPage> startPage;
    shared_ptr<FeedList> feedList;
    QString opmlPath;
    QStringList urlsToImport;
    QHash<QString, FeedEntry> feedInfo;
    QString resourceIdentifier;
};

MigrateFeedsCommand::MigrateFeedsCommand( QObject* parent ) : KJob( parent ), d( new Private( this ) ) {
}

MigrateFeedsCommand::~MigrateFeedsCommand() {
    delete d;
}

void MigrateFeedsCommand::Private::wizardClosed() {
    q->emitResult();
}

void MigrateFeedsCommand::setParentWidget( QWidget* w ) {
    d->parentWidget = w;
}

void MigrateFeedsCommand::start() {
    QMetaObject::invokeMethod( this, "doStart" );
}

void MigrateFeedsCommand::Private::ensureWizardCreated() {
    if ( wizard )
        return;
    wizard = new QWizard( parentWidget );
    wizard->setAttribute( Qt::WA_DeleteOnClose );
    connect( wizard, SIGNAL(destroyed(QObject*)), q, SLOT(wizardClosed()) );
    q->connect( wizard, SIGNAL(currentIdChanged(int)),
                 q, SLOT(currentIdChanged(int)) );
    wizard->setWindowTitle( i18n("Akonadi Data Migration" ) );

    const QStringList resIds = ResourceManager::self()->identifiers();

    StartPage* startPage = new StartPage( resIds.size() > 1 ? resIds : QStringList() );
    startPage->setCommitPage( true );
    wizard->addPage( startPage );
    opmlImportResultPage = new OpmlImportResultPage;
    opmlImportResultPage->setCommitPage( true );
    opmlImportResultPage->setCompleted( false );
    opmlImportResultPageId = wizard->addPage( opmlImportResultPage );
    itemImportResultPage = new ItemImportResultPage;
    itemImportResultPageId = wizard->addPage( itemImportResultPage );
    itemImportResultPage->setCommitPage( true );
}

void MigrateFeedsCommand::Private::ensureWizardShown() {
    ensureWizardCreated();
    wizard->show();
}

void MigrateFeedsCommand::setOpmlFile( const QString& path ) {
    d->opmlPath = path;
}

void MigrateFeedsCommand::Private::doStart() {
    if ( !QFile::exists( opmlPath ) ) {
        q->emitResult();
        return;
    }

    //TODO: check "already migrated" setting

    ensureWizardShown();
}

void MigrateFeedsCommand::Private::resourceCreated( const QString& id ) {
    resourceIdentifier = id;
    assert( !resourceIdentifier.isEmpty() );
    RetrieveFeedListJob* job = new RetrieveFeedListJob( q );
    job->setResourceIdentifiers( QStringList() << resourceIdentifier );
    connect( job, SIGNAL(finished(KJob*)),
             q, SLOT(feedListRetrievalFinished(KJob*)) );
    job->start();
}

void MigrateFeedsCommand::Private::feedListRetrievalFinished( KJob* j ) {
    RetrieveFeedListJob* job = qobject_cast<RetrieveFeedListJob*>( j );
    assert( job );

    if ( job->error() ) {
        //TODO
        return;
    }

    feedList = job->feedList();
    assert( feedList );
    connect( feedList.get(), SIGNAL(feedAdded(KRss::Feed*)), q, SLOT(feedAdded(KRss::Feed*)) );
    startOpmlImport();
}

void MigrateFeedsCommand::Private::feedAdded( Feed* feed ) {
    kDebug() << feed->title();
    const QString xmlUrl = static_cast<const NetFeed*>( feed )->xmlUrl();
    urlsToImport.append( xmlUrl );
    feedInfo[xmlUrl].title = feed->title();
}

void MigrateFeedsCommand::Private::determineResource() {
    const QStringList ids = ResourceManager::self()->identifiers();

    if ( ids.size() == 1 )
        resourceIdentifier = ids.first();
    else if ( startPage->m_resourceBox )
        resourceIdentifier = startPage->m_resourceBox->itemData( startPage->m_resourceBox->currentIndex() ).toString();

    if ( !resourceIdentifier.isEmpty() ) {
        resourceCreated( resourceIdentifier );
        return;
    }

    assert( !"Creating resource not yet implemented. Please create a resource manually." );
}

void MigrateFeedsCommand::Private::opmlImportFinished( KJob* job ) {

    opmlImportResultPage->setState( OpmlImportResultPage::Finished );
    opmlImportResultPage->setCompleted( true );
    if ( job->error() ) {
        opmlImportResultPage->m_resultLabel->setText( i18n("<p>The import of the existing feed list failed:</p><p>%1</p><p>Could not import your feeds and archived articles!</p>", job->errorText() ) );
        return;
    }

    opmlImportResultPage->m_resultLabel->setText( i18n("<p>The existing feed list was successfully imported. In the next step, the archived articles from these feeds will be imported.</p>" ) );
}

void MigrateFeedsCommand::Private::startOpmlImport() {
    const Resource* const resource = ResourceManager::self()->resource( resourceIdentifier );
    assert( resource ); //TODO: remove assertion, handl error
    ImportOpmlJob* job = resource->createImportOpmlJob( KUrl::fromPath( opmlPath ) );
    connect( job, SIGNAL(finished(KJob*)), q, SLOT(opmlImportFinished(KJob*)) );
    job->start();
    opmlImportResultPage->setState( OpmlImportResultPage::Running );
    opmlImportResultPage->m_resultLabel->setText( i18n( "Importing feed list..." ) );
}

void MigrateFeedsCommand::Private::startNextItemImport() {
    if ( urlsToImport.isEmpty() ) {
        itemImportResultPage->m_progressLabel->setText( i18n("Import of archived articles finished.") );
        itemImportResultPage->setState( ItemImportResultPage::Finished );
        itemImportResultPage->setCompleted( true );
        return;
    }

    const Resource* const resource = ResourceManager::self()->resource( resourceIdentifier );
    assert( resource ); //TODO: remove assertion, handl error

    const QString url = urlsToImport.takeFirst();
    Akregator::ImportItemsJob* job = new Akregator::ImportItemsJob( resource, q );
    q->connect( job, SIGNAL(finished(KJob*)), q, SLOT(itemImportFinished(KJob*)) );
    itemImportResultPage->connectJob( job, feedInfo[url].title );
    job->start();
    itemImportResultPage->setState( ItemImportResultPage::Running );
    itemImportResultPage->m_progressLabel->setText( i18nc( "current position, total number, feed title", "Importing items for feed %1 of %2: \"%3\"...", i18n("TODO"), i18n("TODO"), feedInfo[url].title ) );
}

void MigrateFeedsCommand::Private::itemImportFinished( KJob* ) {
    startNextItemImport();
}

void MigrateFeedsCommand::Private::currentIdChanged( int id ) {
    if ( id == opmlImportResultPageId ) {
        determineResource();
        return;
    }
    if ( id == itemImportResultPageId ) {
        startNextItemImport();
        return;
    }
}

#include "migratefeedscommand.moc"
#include "migratefeedswizard_p.moc"
