/*
    Copyright (C) 2008,2009    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "feedlistmodel.h"
#include "tag.h"
#include "tagprovider.h"
#include "treenode.h"
#include "feedlist.h"
#include "aggregatedfeed.h"

#include <KDebug>
#include <KLocale>

#include <QPointer>
#include <QtCore/QHashIterator>

#include <cassert>

#include <boost/shared_ptr.hpp>

using namespace boost;
using namespace KRss;

namespace {
    class SignalBlocker
    {
    public:
        explicit SignalBlocker( QObject* o ) : object( o ), wasBlocked( o->blockSignals( true ) ) {}
        ~SignalBlocker() { if ( object ) object->blockSignals( wasBlocked ); }
    private:
        const QPointer<QObject> object;
        const bool wasBlocked;
    };

    /** helper class, represents a Tag in the FeedListModel, ie.
     a set of feeds under the same tag.
    */
    class AggregatedTagFeed : public AggregatedFeed
    {
    public:
        explicit AggregatedTagFeed( const Tag& tag, QObject *parent = 0 )
            : AggregatedFeed( parent ), m_tag( tag )
        {
        }

        void setTag( const Tag& tag )
        {
            m_tag = tag;
        }
        Tag tag() const
        {
            return m_tag;
        }

        QString title() const
        {
            return m_tag.label();
        }

        void setTitle( const QString &title )
        {
            m_tag.setLabel( title );
        }

        QString description() const
        {
            return m_tag.description();
        }
        void setDescription( const QString &description )
        {
            m_tag.setDescription( description );
        }

    private:
        Tag m_tag;
    };
}


namespace KRss {

class FeedListModelPrivate
{
public:

    explicit FeedListModelPrivate( const shared_ptr<const FeedList>& feedList, const TagProvider *provider,
                                   FeedListModel * const qq );
    ~FeedListModelPrivate();

    TreeNode* addTagNode( const Tag& tag );
    void addFeed( Feed* feed );
    void addFeedNode( Feed *feed, const TagId &tag );
    void removeFeedNode( TreeNode *feedNode );
    QList<TagId> lookupTags( Feed *feed ) const;
    TreeNode* lookupFeedNode( Feed *feed, const TagId &tagId ) const;
    void dump() const;

    void slotTagCreated( const Tag& tag );
    void slotTagModified( const Tag& tag );
    void slotTagDeleted( const TagId& id );
    void slotFeedAdded( KRss::Feed *feed );
    void slotFeedChanged( KRss::Feed *feed );
    void slotFeedRemoved( KRss::Feed *feed );
    void slotFeedItemCountChanged( KRss::Feed *feed, int count );
    void reset();

public:

    FeedListModel * const q;

    shared_ptr<const FeedList> m_feedList;
    const TagProvider* m_tagProvider;
    Tag m_allFeedsTag;

    TreeNode *m_rootNode;
    QHash<Feed*, QList<TreeNode*> > m_feedNodes;
    QHash<TagId, TreeNode*> m_tagNodes;
};

} // namespace KRss

// private implementation
FeedListModelPrivate::FeedListModelPrivate( const shared_ptr<const FeedList>& feedList, const TagProvider *provider,
                                            FeedListModel * const qq )
    : q( qq ), m_feedList( feedList ), m_tagProvider( provider )
{
    // construct the root node
    m_rootNode = new TreeNode( TreeNode::RootNode, 0, 0 );

    // construct the "All feeds" node
    m_allFeedsTag = Tag( TagId( "all-feeds-tag-for-model" ) );
    m_allFeedsTag.setLabel( i18n( "All feeds" ) );
    m_allFeedsTag.setDescription( i18n( "Contains all your feeds" ) );
    addTagNode( m_allFeedsTag );

    // walk over the TagProvider and construct tag nodes for all tags
    const QList<Tag> allTags = m_tagProvider->tags().values();
    Q_FOREACH( const Tag& tag, allTags ) {
        addTagNode( tag );
    }
}

FeedListModelPrivate::~FeedListModelPrivate()
{
    delete m_rootNode;
}

TreeNode* FeedListModelPrivate::addTagNode( const Tag& tag )
{
    Feed *tagFeed = new AggregatedTagFeed( tag );
    TreeNode *tagNode = new TreeNode( TreeNode::TagNode, tagFeed, m_rootNode );

    m_rootNode->appendChild( tagNode );
    m_tagNodes.insert( tag.id(), tagNode );
    return tagNode;
}

void FeedListModelPrivate::addFeed( Feed* feed ) {
    QList<TagId> feedTags = feed->tags();
    if ( feedTags.isEmpty() ) {
        feedTags << m_allFeedsTag.id();
    }

    Q_FOREACH( const TagId &tag, feedTags )
        addFeedNode( feed, tag );
}

void FeedListModelPrivate::addFeedNode( Feed *feed, const TagId &tagId )
{
    Q_ASSERT( m_tagNodes.contains( tagId ) );

    TreeNode *tagNode = m_tagNodes.value( tagId );

    // create the feed node
    QModelIndex tagIndex = q->createIndex( tagNode->row(), 0, tagNode );
    int feedrow = tagNode->count();
    q->beginInsertRows( tagIndex, feedrow, feedrow );
    TreeNode *feedNode = new TreeNode( TreeNode::FeedNode, feed, tagNode );
    tagNode->appendChild( feedNode );
    m_feedNodes[ feed ].append( feedNode );
    q->endInsertRows();

    // add the feed to the aggregated feed inside tagNode
    static_cast<AggregatedFeed*>( tagNode->feed() )->addSourceFeed( feed );
}

void FeedListModelPrivate::removeFeedNode( TreeNode *feedNode )
{
    Feed *feed = feedNode->feed();

    // remove the feed node
    TreeNode *tagNode = feedNode->parent();
    int tagrow = tagNode->row();
    QModelIndex tagIndex = q->createIndex( tagrow, 0, tagNode );
    int feedrow = feedNode->row();
    q->beginRemoveRows( tagIndex, feedrow, feedrow );
    m_feedNodes[ feed ].removeOne( feedNode );
    tagNode->removeChildAt( feedrow );
    q->endRemoveRows();

    // remove the feed from the aggregated feed inside tagNode
    static_cast<AggregatedFeed*>( tagNode->feed() )->removeSourceFeed( feed );
}

QList<TagId> FeedListModelPrivate::lookupTags( Feed *feed ) const
{
    QList<TreeNode*> feedNodes = m_feedNodes.value( feed );
    QList<TagId> tags;
    Q_FOREACH( const TreeNode *feedNode, feedNodes ) {
        const Tag& tag = static_cast<const AggregatedTagFeed*>( feedNode->parent()->feed() )->tag();
        tags.append( tag.id() );
    }

    return tags;
}

TreeNode* FeedListModelPrivate::lookupFeedNode( Feed *feed, const TagId &tag ) const
{
    TreeNode *tagNode = m_tagNodes.value( tag );
    QList<TreeNode*> feedNodes = m_feedNodes.value( feed );
    Q_FOREACH( TreeNode *feedNode, feedNodes ) {
        if ( feedNode->parent() == tagNode )
            return feedNode;
    }

    return 0;
}

void FeedListModelPrivate::dump() const
{
    kDebug() << "Feed nodes by ids:";
    QHashIterator<Feed*, QList<TreeNode*> > it_feed( m_feedNodes );
    while ( it_feed.hasNext() ) {
        it_feed.next();
        kDebug() << "Feed id:" << it_feed.key()->id() << ", feed nodes:" << it_feed.value();
    }

    kDebug() << "Tag nodes by tags:";
    QHashIterator<TagId, TreeNode*> it_tag( m_tagNodes );
    while ( it_tag.hasNext() ) {
        it_tag.next();
        kDebug() << "Tag:" << Tag::idToString( it_tag.key() ) << ", tag nodes: " << it_tag.value();
    }

    m_rootNode->dump();
}

void FeedListModelPrivate::reset()
{
    // we block signals and then call reset() to update
    // the model in one sitting
    if ( m_feedList ) {
        const SignalBlocker blocker( q );
        Q_FOREACH( Feed * const feed, m_feedList->feeds() )
            addFeed( feed );
    }
    q->reset();

    dump();
}

void FeedListModelPrivate::slotTagCreated( const KRss::Tag& tag )
{
    const TreeNode *tagNode = addTagNode( tag );
    const int row = tagNode->row();
    q->beginInsertRows( QModelIndex(), row, row );
    q->endInsertRows();
}

void FeedListModelPrivate::slotTagModified( const KRss::Tag& tag )
{
    Q_ASSERT( m_tagNodes.contains( tag.id() ) );

    TreeNode *tagNode = m_tagNodes.value( tag.id() );
    const QModelIndex tagIndex = q->createIndex( tagNode->row(), 0, tagNode );
    static_cast<AggregatedTagFeed*>( tagNode->feed() )->setTag( tag );
    emit q->dataChanged( tagIndex, tagIndex );
}

void FeedListModelPrivate::slotTagDeleted( const KRss::TagId& id )
{
    Q_ASSERT( m_tagNodes.value( id )->count() == 0 );

    const TreeNode *tagNode = m_tagNodes.value( id );
    const int row = tagNode->row();
    q->beginRemoveRows( QModelIndex(), row, row );
    m_tagNodes.remove( id );
    delete tagNode->feed();
    m_rootNode->removeChildAt( row );
    q->endRemoveRows();
}

void FeedListModelPrivate::slotFeedAdded( Feed *feed )
{
    addFeed( feed );
}

void FeedListModelPrivate::slotFeedChanged( Feed *feed )
{
    QList<TagId> newTags = feed->tags();
    if ( newTags.isEmpty() ) {
        newTags << m_allFeedsTag.id();
    }

    QList<TagId> oldTags = lookupTags( feed );
    kDebug() << "New tags:" << newTags;
    kDebug() << "Old tags:" << oldTags;
    QSet<TagId> common = oldTags.toSet().intersect( newTags.toSet() );
    QSet<TagId> toRemove = oldTags.toSet().subtract( newTags.toSet() );
    QSet<TagId> toAdd = newTags.toSet().subtract( oldTags.toSet() );
    kDebug() << "common:" << common;
    kDebug() << "to remove:" << toRemove;
    kDebug() << "to add:" << toAdd;

    // check whether this feed has no tags now
    if ( common.count() == 0 && toAdd.count() == 0 ) {
        toAdd << m_allFeedsTag.id();
    }

    // updated the feed nodes under the 'common' tags
    Q_FOREACH( const TagId &tag, common ) {
        TreeNode *feedNode = lookupFeedNode( feed, tag );
        QModelIndex feedIndex = q->createIndex( feedNode->row(), 0, feedNode );
        emit q->dataChanged( feedIndex, feedIndex );
    }

    // remove the feed nodes from the 'toRemove' tags
    Q_FOREACH( const TagId &tag, toRemove ) {
        TreeNode *feedNode = lookupFeedNode( feed, tag );
        removeFeedNode( feedNode );
    }

    // add new feed nodes under the 'toAdd' tags
    Q_FOREACH( const TagId &tag, toAdd ) {
        addFeedNode( feed, tag );
    }

    dump();
}

void FeedListModelPrivate::slotFeedItemCountChanged( KRss::Feed *feed, int ) {
    //TODO: make this more efficient
    slotFeedChanged( feed );
}

void FeedListModelPrivate::slotFeedRemoved( Feed *feed )
{
    QList<TreeNode*> feedNodes = m_feedNodes.value( feed );
    Q_FOREACH( TreeNode *feedNode, feedNodes ) {
        removeFeedNode( feedNode );
    }

    dump();
}

// public interface implementation
FeedListModel::FeedListModel( const shared_ptr<const FeedList>& feedList, const TagProvider *provider, QObject *parent )
    : QAbstractItemModel( parent ), d( new FeedListModelPrivate( feedList, provider, this ) )
{
    if ( d->m_feedList ) {
        connect( d->m_feedList.get(), SIGNAL( feedAdded( KRss::Feed* ) ),
                    this, SLOT( slotFeedAdded( KRss::Feed* ) ) );
        connect( d->m_feedList.get(), SIGNAL( feedChanged( KRss::Feed* ) ),
                    this, SLOT( slotFeedChanged( KRss::Feed* ) ) );
        connect( d->m_feedList.get(), SIGNAL( feedRemoved( KRss::Feed* ) ),
                    this, SLOT( slotFeedRemoved( KRss::Feed* ) ) );
        connect( d->m_feedList.get(), SIGNAL( unreadCountChanged( KRss::Feed*,int ) ),
                    this, SLOT( slotFeedItemCountChanged( KRss::Feed*, int ) ) );
        connect( d->m_feedList.get(), SIGNAL( totalCountChanged( KRss::Feed*,int ) ),
                    this, SLOT( slotFeedItemCountChanged( KRss::Feed*, int ) ) );

        Q_FOREACH( Feed* const feed, d->m_feedList->feeds() )
            d->addFeed( feed );
    }

    connect( d->m_tagProvider, SIGNAL( tagCreated( const KRss::Tag& ) ),
             this, SLOT( slotTagCreated( const KRss::Tag& ) ) );
    connect( d->m_tagProvider, SIGNAL( tagModified( const KRss::Tag& ) ),
             this, SLOT( slotTagModified( const KRss::Tag& ) ) );
    connect( d->m_tagProvider, SIGNAL( tagDeleted( const KRss::TagId& ) ),
             this, SLOT( slotTagDeleted( const KRss::TagId& ) ) );
}

FeedListModel::~FeedListModel()
{
    delete d;
}

shared_ptr<const FeedList> FeedListModel::feedList() const
{
    return d->m_feedList;
}

QVariant FeedListModel::headerData( int section, Qt::Orientation orientation, int role ) const
{
    if ( section == 0 && orientation == Qt::Horizontal && role == Qt::DisplayRole )
        return i18n( "Feeds" );

    return QVariant();
}

int FeedListModel::columnCount( const QModelIndex &parent ) const
{
    Q_UNUSED( parent );
    return 1;
}

int FeedListModel::rowCount( const QModelIndex &parent ) const
{
    if ( !parent.isValid() )
        return d->m_rootNode->count();

    TreeNode *node = static_cast<TreeNode*>( parent.internalPointer() );
    return node->count();
}

QVariant FeedListModel::data(const QModelIndex &index, int role ) const
{
    if ( !index.isValid() )
        return QVariant();

    TreeNode *node = static_cast<TreeNode*>( index.internalPointer() );

    QVariant var;
    switch ( role ) {
        case Qt::DisplayRole:
        {
            const int unread = node->feed()->unread();
            if ( unread > 0 )
                return i18nc( "feed title (unread count)", "%1 (%2)", node->feed()->title(), unread );
            else
                return node->feed()->title();
        }
        case Qt::ToolTipRole:       return node->feed()->description();
        case TierRole:              return ( node->tier() == TreeNode::TagNode ? TagTier : FeedTier );
        case FeedIdRole:            return node->feed()->id();
        case FeedRole:
                                    var.setValue( node->feed() );
                                    return var;
        case TagIdRole:
                                    return static_cast<const AggregatedTagFeed*>( node->feed() )->tag().id();
        case TagRole:
                                    var.setValue( static_cast<const AggregatedTagFeed*>( node->feed() )->tag() );
                                    return var;
        default:                    return QVariant();
    }

    return QVariant();
}

QModelIndex FeedListModel::index( int row, int column, const QModelIndex &parent ) const
{
    if ( row < 0 || column < 0 )
        return QModelIndex();

    // create index for a tag node
    if ( !parent.isValid() ) {
        if ( row < d->m_rootNode->count() && column == 0 ) {
            return createIndex( row, 0, d->m_rootNode->childAt( row ) );
        }
        else {
            return QModelIndex();
        }
    }

    TreeNode *node = static_cast<TreeNode*>( parent.internalPointer() );
    if ( row < node->count() && column == 0 ) {
        TreeNode *child = node->childAt( row );
        return createIndex( row, 0, child );
    }
    else {
        return QModelIndex();
    }
}

QModelIndex FeedListModel::parent( const QModelIndex &index ) const
{
    if ( !index.isValid() )
        return QModelIndex();

    TreeNode *node = static_cast<TreeNode*>( index.internalPointer() );

    if ( node->tier() == TreeNode::TagNode ) {
        return QModelIndex();
    }
    else if ( node->tier() == TreeNode::FeedNode ) {
        TreeNode *parentNode = node->parent();
        return createIndex( d->m_rootNode->childRow( parentNode ), 0, parentNode );
    }

    return QModelIndex();
}

QList<QModelIndex> FeedListModel::feedIndexes( Feed *feed ) const
{
    QList<QModelIndex> feedIndexesList;

    QList<TreeNode*> feedNodes = d->m_feedNodes.value( feed );
    Q_FOREACH( TreeNode *feedNode, feedNodes ) {
        feedIndexesList << createIndex ( feedNode->row(), 0, feedNode );
    }

    return feedIndexesList;
}

#include "feedlistmodel.moc"
