/*
 * This file is part of the krss library
 *
 * Copyright (C) 2009 Frank Osterfeld <osterfeld@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "importopmljob.h"
#include "krssinterface.h"
#include "resource.h"

#include <kio/job.h>
#include <kio/jobclasses.h>

#include <KLocalizedString>
#include <KTemporaryFile>
#include <KUrl>

#include <cerrno>
#include <cassert>

using namespace KIO;
using namespace KRss;

class ImportOpmlJob::Private {
    ImportOpmlJob* const q;
public:
    explicit Private( const QString& dbusService, ImportOpmlJob* q );

    void doStart();
    void downloadFinished( KJob* );
    void importFinished( bool );
    void startImport( const QString& path );

    KUrl url;
    KTemporaryFile tempFile;
    org::kde::krss* interface;
};

ImportOpmlJob::Private::Private( const QString& dbusService, ImportOpmlJob* qq )
  : q( qq )
  , interface( new org::kde::krss( dbusService,
                                   "/KRss",
                                    QDBusConnection::sessionBus(),
                                    q ) )
{
}


ImportOpmlJob::ImportOpmlJob( const QString& dbusService, QObject* parent ) : KJob( parent ), d( new Private( dbusService, this ) ) {

}

ImportOpmlJob::~ImportOpmlJob() {
    delete d;
}

void ImportOpmlJob::setSourceUrl( const KUrl& url ) {
    d->url = url;
}

void ImportOpmlJob::start() {
    QMetaObject::invokeMethod( this, "doStart", Qt::QueuedConnection );
}

void ImportOpmlJob::Private::doStart() {
    if ( url.isLocalFile() ) {
        startImport( url.path() );
        return;
    }
    StoredTransferJob* const job = KIO::storedGet( url );
    q->connect( job, SIGNAL(finished(KJob*)),
                q, SLOT(downloadFinished(KJob*)) );
    job->start();
}

void ImportOpmlJob::Private::startImport( const QString& path ) {
    if ( !interface->callWithCallback( "importOpml", QList<QVariant>() << path << QString(), q, SLOT(importFinished(bool)) ) ) {
        q->setError( ImportOpmlJob::ImportFailedError );
        q->setErrorText( i18n( "Could not start import." ) );
        q->emitResult();
    }
}

void ImportOpmlJob::Private::importFinished( bool success ) {
    if ( !success ) {
        q->setError( ImportOpmlJob::ImportFailedError );
        q->setErrorText( i18n( "Import into RSS resource failed." ) );
    }
    q->emitResult();
}

void ImportOpmlJob::Private::downloadFinished( KJob* j ) {
    const StoredTransferJob* const job = qobject_cast<StoredTransferJob*>( j );
    assert( j );
    if ( job->error() ) {
        q->setError( ImportOpmlJob::FileAccessError );
        q->setErrorText( job->errorText() );
        q->emitResult();
        return;
    }

    const QString writeError = i18n( "Could not write downloaded OPML file to temporary file." );

    const QByteArray data = job->data();
    if ( !tempFile.open() ) {
        q->setError( ImportOpmlJob::FileAccessError );
        q->setErrorText( writeError );
        q->emitResult();
        return;
    }

    errno = 0;
    qint64 toWrite = data.size();
    qint64 written = 0;
    while ( toWrite > 0 ) {
        const qint64 num = tempFile.write( data.constData() + written, toWrite );
        if ( num < 0 ) {
            q->setError( ImportOpmlJob::FileAccessError );
            q->setErrorText( writeError );
            q->emitResult();
            return;
        }
        written += num;
        toWrite -= num;
    }

    tempFile.close();

    startImport( tempFile.fileName() );
}

#include "importopmljob.moc"
