/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "persistentfeed.h"
#include "persistentfeed_p.h"
#include "item_p.h"
#include "resource.h"

#include <akonadi/itemfetchjob.h>
#include <akonadi/collectionfetchjob.h>
#include <akonadi/collectionstatistics.h>
#include <akonadi/collectionstatisticsjob.h>
#include <akonadi/item.h>
#include <akonadi/itemfetchscope.h>
#include <KIcon>
#include <KDebug>

#include <cassert>

using Akonadi::Collection;
using Akonadi::ItemFetchJob;
using Akonadi::CollectionFetchJob;

using namespace KRss;

PersistentFeed::PersistentFeed( const FeedCollection &feedCollection, const Resource *resource, QObject *parent )
    : Feed( parent ), d( new PersistentFeedPrivate( feedCollection, resource, this ) )
{
    connect( d->m_resource, SIGNAL( feedChanged( KRss::Feed::Id ) ),
             this, SLOT( slotFeedChanged( KRss::Feed::Id ) ) );
    connect( d->m_resource, SIGNAL( feedRemoved( KRss::Feed::Id ) ),
             this, SLOT( slotFeedRemoved( KRss::Feed::Id ) ) );

    connect( d->m_resource, SIGNAL( fetchStarted( KRss::Feed::Id ) ),
             this, SLOT( slotFetchStarted( KRss::Feed::Id ) ) );
    connect( d->m_resource, SIGNAL( fetchPercent( KRss::Feed::Id, uint ) ),
             this, SLOT( slotFetchPercent( KRss::Feed::Id, uint ) ) );
    connect( d->m_resource, SIGNAL( fetchFinished( KRss::Feed::Id ) ),
             this, SLOT( slotFetchFinished( KRss::Feed::Id ) ) );
    connect( d->m_resource, SIGNAL( fetchFailed( KRss::Feed::Id, const QString& ) ),
             this, SLOT( slotFetchFailed( KRss::Feed::Id, const QString& ) ) );
    connect( d->m_resource, SIGNAL( statisticsChanged(KRss::Feed::Id, Akonadi::CollectionStatistics) ),
             this , SLOT( slotStatisticsChanged(KRss::Feed::Id, Akonadi::CollectionStatistics)) );

    connect( d->m_resource, SIGNAL( itemAdded( const Akonadi::Item&, KRss::Feed::Id ) ),
             this, SLOT( slotItemAdded( const Akonadi::Item&, KRss::Feed::Id ) ) );
    connect( d->m_resource, SIGNAL( itemChanged( const Akonadi::Item& ) ),
             this, SLOT( slotItemChanged( const Akonadi::Item& ) ) );
    connect( d->m_resource, SIGNAL( itemRemoved( const Akonadi::Item& ) ),
             this, SLOT( slotItemRemoved( const Akonadi::Item& ) ) );

    Akonadi::CollectionStatisticsJob* job = new Akonadi::CollectionStatisticsJob( feedCollection );
    connect( job, SIGNAL(finished(KJob*)), this, SLOT(slotStatisticsFetchDone(KJob*)) );
}

PersistentFeed::~PersistentFeed()
{
    delete d;
}

Feed::Id PersistentFeed::id() const
{
    return d->m_feedCollection.id();
}

QString PersistentFeed::title() const
{
    return d->m_feedCollection.title();
}

void PersistentFeed::setTitle( const QString &title )
{
    d->m_feedCollection.setTitle( title );
}

QString PersistentFeed::description() const
{
    return d->m_feedCollection.description();
}

void PersistentFeed::setDescription( const QString &description )
{
    d->m_feedCollection.setDescription( description );
}

QString PersistentFeed::type() const
{
    return d->m_feedCollection.feedType();
}

void PersistentFeed::setType( const QString &type )
{
    d->m_feedCollection.setFeedType( type );
}

KIcon PersistentFeed::icon() const
{
    return KIcon();
}

void PersistentFeed::setIcon( const KIcon &icon )
{
    Q_UNUSED( icon )
}

QList<TagId> PersistentFeed::tags() const
{
    return d->m_feedCollection.tags();
}

void PersistentFeed::setTags( const QList<TagId> &tags )
{
    d->m_feedCollection.setTags( tags );
}

void PersistentFeed::addTag( const TagId &tag )
{
    d->m_feedCollection.addTag( tag );
}

void PersistentFeed::removeTag( const TagId &tag )
{
    d->m_feedCollection.removeTag( tag );
}

QPixmap PersistentFeed::image() const
{
    return QPixmap();
}

void PersistentFeed::setImage( const QPixmap &image )
{
    Q_UNUSED( image )
}

QStringList PersistentFeed::subscriptionLabels() const
{
    return d->m_feedCollection.subscriptionLabels();
}

void PersistentFeed::setSubscriptionLabels( const QStringList &subscriptionLabels )
{
    d->m_feedCollection.setSubscriptionLabels( subscriptionLabels );
}

void PersistentFeed::addSubscriptionLabel( const QString &subscriptionLabel )
{
    d->m_feedCollection.addSubscriptionLabel( subscriptionLabel );
}

void PersistentFeed::removeSubscriptionLabel( const QString &subscriptionLabel )
{
    d->m_feedCollection.removeSubscriptionLabel( subscriptionLabel );
}

int PersistentFeed::total() const
{
    return d->m_totalCount;
}

int PersistentFeed::unread() const
{
    return d->m_unreadCount;
}

int PersistentFeed::fetchInterval() const
{
    return ( d->m_feedCollection.fetchInterval() > 0 ? d->m_feedCollection.fetchInterval() : 0 );
}

void PersistentFeed::setFetchInterval( int minutes )
{
    d->m_feedCollection.setFetchInterval( minutes > 0 ? minutes : -1 );
}

int PersistentFeed::maxArticleNumber() const
{
    return 0;
}

void PersistentFeed::setMaxArticleNumber( int number )
{
    Q_UNUSED( number )
}

int PersistentFeed::maxArticleAge() const
{
    return 0;
}

void PersistentFeed::setMaxArticleAge( int age )
{
    Q_UNUSED( age )
}

void PersistentFeedItemListJobImpl::jobDone( KJob* j ) {
    assert( m_job == j );
    setError( m_job->error() );
    setErrorText( m_job->errorString() );
    if ( m_job->error() ) {
        emitResult();
        return;
    }

    Q_FOREACH( const Akonadi::Item &loadedItem, m_job->items() ) {
        Item item;
        item.d->akonadiItem = loadedItem;
        m_items.append( item );
    }

    emit result( m_items );
    emitResult();
}

void PersistentFeedItemListJobImpl::start() {
    ItemFetchJob *job = new ItemFetchJob( m_feed->feedCollection(), this );
    if ( !m_fetchScope.isEmpty() )
        job->setFetchScope( m_fetchScope );
    connect(job, SIGNAL(finished(KJob*)), this, SLOT(jobDone(KJob*)) );
    m_job = job;
    job->start();
}

ItemListJob* PersistentFeed::createListJob() {
    return new PersistentFeedItemListJobImpl( this );
}


void PersistentFeed::fetch() const
{
    if ( d->m_resource )
        d->m_resource->fetchFeed( d->m_feedCollection.id() );
}

void PersistentFeed::abortFetch() const
{
    if ( d->m_resource )
        d->m_resource->abortFetch( d->m_feedCollection.id() );
}

FeedCollection& PersistentFeed::feedCollection() const
{
    return d->m_feedCollection;
}

void PersistentFeed::slotCollectionLoadDone( KJob *job )
{
    if ( job->error() ) {
        kWarning() << "Collection load failed:" << job->errorString();
        return;
    }

    d->m_feedCollection = static_cast<CollectionFetchJob*>( job )->collections().first();
    emit changed( this );
}

void PersistentFeed::slotFeedChanged( Feed::Id id )
{
    if ( d->m_feedCollection.id() == id ) {
        CollectionFetchJob *job = new CollectionFetchJob( d->m_feedCollection, CollectionFetchJob::Base, this );
        connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotCollectionLoadDone( KJob* ) ) );
    }
}

void PersistentFeed::slotFeedRemoved( Feed::Id id )
{
    if ( d->m_feedCollection.id() == id )
        emit removed( this );
}

void slotStatisticsFetchDone( KJob* job );

void PersistentFeedPrivate::updateFromStatistics( const Akonadi::CollectionStatistics& stats ) {
    if ( m_unreadCount != stats.unreadCount() ) {
        m_unreadCount = stats.unreadCount();
        emit q->unreadCountChanged( q, m_unreadCount );
    }
    if ( m_totalCount != stats.count() ) {
        m_totalCount = stats.count();
        emit q->totalCountChanged( q, m_totalCount );
    }
}

void PersistentFeed::slotStatisticsFetchDone( KJob* j )
{
    const Akonadi::CollectionStatisticsJob* const job = qobject_cast<Akonadi::CollectionStatisticsJob*>( j );
    assert( job );
    if ( job->error() ) {
        kWarning() << "Fetching statistics failed. Error:" << job->error() << job->errorText();
        return;
    }
    d->updateFromStatistics( job->statistics() );
}

void PersistentFeed::slotStatisticsChanged( KRss::Feed::Id id, const Akonadi::CollectionStatistics& stats )
{
    kDebug() << id << stats.unreadCount();
    if ( d->m_feedCollection.id() != id )
        return;
    d->updateFromStatistics( stats );

}

void PersistentFeed::slotFetchStarted( Feed::Id id )
{
    if ( d->m_feedCollection.id() == id )
        emit fetchStarted( this );
}

void PersistentFeed::slotFetchPercent( Feed::Id id, uint percentage )
{
    if ( d->m_feedCollection.id() == id )
        emit fetchPercent( this, percentage );
}

void PersistentFeed::slotFetchFinished( Feed::Id id )
{
    if ( d->m_feedCollection.id() == id )
        emit fetchFinished( this );
}

void PersistentFeed::slotFetchFailed( Feed::Id id, const QString &errorMessage )
{
    if ( d->m_feedCollection.id() == id )
        emit fetchFailed( this, errorMessage );
}

void PersistentFeed::slotFetchAborted( Feed::Id id )
{
    if ( d->m_feedCollection.id() == id )
        emit fetchAborted( this );
}

void PersistentFeed::slotItemAdded( const Akonadi::Item &item, Feed::Id id )
{
    Item it;
    it.d->akonadiItem = item;
    emit itemsAdded( this, QList<Item>() << it );
}

void PersistentFeed::slotItemChanged( const Akonadi::Item &item )
{
    Item it;
    it.d->akonadiItem = item;
    emit itemsChanged( this, QList<Item>() << it );
}

void PersistentFeed::slotItemRemoved( const Akonadi::Item &item )
{
    Item it;
    it.d->akonadiItem =  item;
    emit itemsRemoved( this, QList<Item>() << it );
}

#include "persistentfeed.moc"
#include "persistentfeed_p.moc"
