/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "fetchfeedjob.h"
#include "flagsmodifyjob.h"
#include "rssitemsync.h"

#include "libkrss/rssitem.h"
#include "libkrss/item.h"
#include "libkrss/person.h"
#include "libkrss/enclosure.h"
#include "libkrss/category.h"

#include <KUrl>
#include <KDebug>
#include <KDateTime>
#include <KLocale>
#include <syndication/dataretriever.h>

#include <QtCore/QTimer>

using namespace KRss;

// from akregator/src/utils.cpp
static inline int calcHash( const QString &str )
{
    const QByteArray array = str.toAscii();
    return qChecksum( array.constData(), array.size() );
}

static inline KRss::RssItem convertToRssItem( const Syndication::ItemPtr &syndItem )
{
    RssItem rssItem;

    rssItem.setStatus( KRss::New );
    rssItem.setGuidIsHash( syndItem->id().startsWith( "hash:" ) );
    //rssItem.setSourceFeedId(currentCollection().remoteId().toInt()); // kfeed supports only integer ids
    rssItem.setTitle( syndItem->title() );
    rssItem.setLink( syndItem->link() );
    rssItem.setDescription( syndItem->description() );
    rssItem.setContent( syndItem->content() );
    KDateTime dt;
    if ( syndItem->datePublished() > 0 ) {
        dt.setTime_t( syndItem->datePublished() );
        rssItem.setDatePublished( dt );
    }
    else {
        rssItem.setDatePublished( KDateTime::currentLocalDateTime() );
    }
    dt.setTime_t( syndItem->dateUpdated() );
    rssItem.setDateUpdated( dt );
    rssItem.setGuid( syndItem->id() );
    rssItem.setLanguage( syndItem->language() );
    rssItem.setCommentsCount( syndItem->commentsCount() );
    rssItem.setCommentsLink( syndItem->commentsLink() );
    rssItem.setCommentsFeed( syndItem->commentsFeed() );
    rssItem.setCommentPostUri( syndItem->commentPostUri() );

    // Authors
    QString authors;  // for hash
    QList<KRss::Person> kfeedPersons;
    Q_FOREACH( const Syndication::PersonPtr &pers, syndItem->authors() ) {
        if ( !pers->isNull() ) {
            KRss::Person kfeedPerson;
            kfeedPerson.setName( pers->name() );
            kfeedPerson.setUri( pers->uri() );
            kfeedPerson.setEmail( pers->email() );
            kfeedPersons << kfeedPerson;
        }
    }
    rssItem.setAuthors( kfeedPersons );

    // Enclosures
    QList<KRss::Enclosure> kfeedEnclosures;
    Q_FOREACH( const Syndication::EnclosurePtr &encl, syndItem->enclosures() ) {
        if ( !encl->isNull() ) {
            KRss::Enclosure kfeedEnclosure;
            kfeedEnclosure.setUrl( encl->url() );
            kfeedEnclosure.setTitle(encl->title());
            kfeedEnclosure.setType( encl->type() );
            kfeedEnclosure.setLength( encl->length() );
            kfeedEnclosure.setDuration( encl->duration() );
            kfeedEnclosures << kfeedEnclosure;
        }
    }
    rssItem.setEnclosures( kfeedEnclosures );

    // Categories
    QList<KRss::Category> kfeedCategories;
    Q_FOREACH( const Syndication::CategoryPtr &cat, syndItem->categories() ) {
        if ( !cat->isNull() ) {
            KRss::Category kfeedCategory;
            kfeedCategory.setTerm( cat->term() );
            kfeedCategory.setScheme( cat->scheme() );
            kfeedCategory.setLabel( cat->label() );
            kfeedCategories << kfeedCategory;
        }
    }
    rssItem.setCategories( kfeedCategories );

    // TODO: custom properties

    rssItem.setHash( calcHash( syndItem->title() + syndItem->description() + syndItem->content() + syndItem->link() +
                       authors ) );
    return rssItem;
}

// private implementation
class FetchFeedJobPrivate
{
public:

    explicit FetchFeedJobPrivate( FetchFeedJob * const qq, const FeedCollection &feedCollection )
        : q( qq ), m_feed( feedCollection ), m_loader( 0 ), m_aborted( false )
    {
    }

    void doStart();
    void slotClearNewDone( KJob* );
    void slotFetchDone( Syndication::Loader *loader, Syndication::FeedPtr feed, Syndication::ErrorCode status );
    void slotSyncDone( KJob *job );
    void slotProgress( unsigned long percentage );

public:

    FetchFeedJob * const q;
    const FeedCollection m_feed;
    Syndication::Loader *m_loader;
    bool m_aborted;
};

void FetchFeedJobPrivate::doStart()
{
    // don't proceed further if aborted
    if ( m_aborted ) {
        q->setError( FetchFeedJob::Aborted );
        q->emitResult();
        return;
    }

    FlagsModifyJob *job = new FlagsModifyJob( m_feed, q );
    job->clearFlags( Akonadi::Item::Flags() << "\\Recent" );
    QObject::connect( job, SIGNAL( result( KJob* ) ), q, SLOT( slotClearNewDone( KJob* ) ) );
}

void FetchFeedJobPrivate::slotClearNewDone( KJob *job )
{
    if ( job->error() ) {
        kWarning() << "Failed to clear the 'New' flag on the items";
        kWarning() << "Loading feed anyway";
    }

    // don't proceed further if aborted
    if ( m_aborted ) {
        q->setError( FetchFeedJob::Aborted );
        q->emitResult();
        return;
    }

    Syndication::Loader *loader = Syndication::Loader::create( q, SLOT( slotFetchDone( Syndication::Loader*,
                                                                                           Syndication::FeedPtr,
                                                                                           Syndication::ErrorCode ) ) );

    kDebug() << "Url:" << m_feed.xmlUrl();
    Syndication::DataRetriever *dr = new Syndication::FileRetriever();
    QObject::connect( dr, SIGNAL( percent( unsigned long ) ), q, SLOT( slotProgress( unsigned long ) ) );
    loader->loadFrom( KUrl( m_feed.xmlUrl() ), dr );
    m_loader = loader;
}

void FetchFeedJobPrivate::slotFetchDone( Syndication::Loader *loader, Syndication::FeedPtr feed, Syndication::ErrorCode status )
{
    Q_UNUSED( loader )

    kDebug() << "Status:" << status;
    m_loader = 0;

    if ( status != Syndication::Success ) {
        q->setError( FetchFeedJob::Aborted -1 + status );
        q->emitResult();
        return;
    }

    // don't proceed further if aborted
    if ( m_aborted ) {
        q->setError( FetchFeedJob::Aborted );
        q->emitResult();
        return;
    }

    Akonadi::Item::List items;
    Q_FOREACH( const Syndication::ItemPtr &it, feed->items() ) {
        Akonadi::Item item;
        item.setRemoteId( it->id() );
        item.setMimeType( "application/rss+xml" );
        item.setPayload<KRss::RssItem>( convertToRssItem( it ) );
        items.append( item );
    }

    Akonadi::ItemSync *syncer = new RssItemSync( m_feed, q );
    QObject::connect( syncer, SIGNAL( result( KJob* ) ), q, SLOT( slotSyncDone( KJob* ) ) );
    syncer->setIncrementalSyncItems( items, Akonadi::Item::List() );
}

void FetchFeedJobPrivate::slotSyncDone( KJob *job )
{
    // don't proceed further if aborted
    if ( m_aborted ) {
        q->setError( FetchFeedJob::Aborted );
        q->emitResult();
        return;
    }

    if ( job->error() ) {
        q->setError( FetchFeedJob::ItemSyncFailed );
        q->setErrorText( job->errorString() );
    }

    q->emitResult();
}

void FetchFeedJobPrivate::slotProgress( unsigned long percentage )
{
    emit q->progress( m_feed.id(), percentage );
}

// public interface implementation
FetchFeedJob::FetchFeedJob( const FeedCollection &feed, QObject *parent ) :
    KJob( parent ), d( new FetchFeedJobPrivate( this, feed ) )
{
}

FetchFeedJob::~FetchFeedJob()
{
    delete d;
}

FeedCollection FetchFeedJob::feed() const
{
    return d->m_feed;
}

void FetchFeedJob::abort()
{
    d->m_aborted = true;

    // Akonadi jobs are not killable, we can only kill Syndication::Loader
    if ( d->m_loader ) {
        d->m_loader->abort();
    }
}

void FetchFeedJob::start()
{
    QTimer::singleShot( 0, this, SLOT( doStart() ) );
}

QString FetchFeedJob::errorString() const
{
    switch ( error() ) {
        // Error strings shamelessly taken from Syndication API docs
        case Aborted:               return i18n( "File downloading/parsing was aborted." );
        case Timeout:               return i18n( "File download timed out." );
        case UnknownHost:           return i18n( "The hostname couldn't get resolved to an IP address." );
        case FileNotFound:          return i18n( "The host was contacted successfully, but reported a 404 error." );
        case OtherRetrieverError:   return i18n( "Retriever error not covered by the error codes above." );
        case InvalidXml:            return i18n( "The XML is invalid." );
        case XmlNotAccepted:        return i18n( "The source is valid XML, but no parser accepted it." );
        case InvalidFormat:         return i18n( "The source was accepted by a parser, but the actual parsing failed." );
        case ItemSyncFailed:        return errorText();
        default:                    return KJob::errorString();
    }
}

#include "fetchfeedjob.moc"
