/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "rssresource.h"
#include "krssadaptor.h"
#include "addfeedjob.h"
#include "fetchfeedjob.h"
#include "exportopmljob.h"
#include "importopmljob.h"
#include "importitemsjob.h"
#include "libkrss/resourcemanager.h"
#include "libkrss/tagprovider.h"

#include <akonadi/collectionfetchjob.h>
#include <akonadi/collectionmodifyjob.h>
#include <akonadi/collectiondeletejob.h>
#include <akonadi/changerecorder.h>

#include <KDebug>
#include <KFileDialog>
#include <KLocale>
#include <KRandom>

#include <QtDBus/QDBusConnection>

using namespace KRss;

// member functions
RssResource::RssResource( const QString &id )
        :ResourceBase( id )
{
    KRss::ResourceManager::registerAttributes();

    changeRecorder()->fetchCollection( true );

    new KrssAdaptor( this );
    if ( !QDBusConnection::sessionBus().registerObject( QLatin1String( "/KRss" ), this, QDBusConnection::ExportAdaptors ) ) {
        kWarning() << "Couldn't register a D-Bus service org.kde.krss";
        kWarning() << QDBusConnection::sessionBus().lastError().message();
    }

    QTimer::singleShot( 0, this, SLOT( initialize() ) );
}

void RssResource::initialize()
{
    TagProviderRetrieveJob *job = new TagProviderRetrieveJob( this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotTagProviderRetrievalDone( KJob* ) ) );
    job->start();
}

void RssResource::collectionAdded( const Akonadi::Collection &collection, const Akonadi::Collection &parent )
{
    Q_ASSERT( parent.id() == m_rootCollection.id() );

    kDebug() << "Collection id:" << collection.id() << " , name:" << collection.name();
    kDebug() << "Adding new feeds outside the RSS resource is not supported!";
    m_feeds[ collection.id() ] = collection;
    changeCommitted( collection );
}

void RssResource::collectionChanged( const Akonadi::Collection &collection )
{
    Q_ASSERT( m_feeds.contains( collection.id() ) );

    kDebug() << "Collection id:" << collection.id() << " , name:" << collection.name();
    m_feeds[ collection.id() ] = collection;
    changeCommitted( collection );
}

void RssResource::collectionRemoved( const Akonadi::Collection &collection )
{
    Q_ASSERT( m_feeds.contains( collection.id() ) );
    // also check if the feed is being processed atm

    kDebug() << "Collection id:" << collection.id() << " , name:" << collection.name();
    abortFetch( collection.id() );
    m_feeds.remove( collection.id() );
    changeCommitted( collection );
}

void RssResource::itemAdded( const Akonadi::Item &item, const Akonadi::Collection &collection )
{
    Q_UNUSED( collection )

    kDebug() << "Item id:" << item.id();
    kDebug() << "Adding new RSS items is not supported!";
    changeCommitted( item );
}

void RssResource::itemChanged( const Akonadi::Item &item, const QSet<QByteArray> &partIdentifiers )
{
    Q_UNUSED( partIdentifiers )

    kDebug() << "Item id:" << item.id();
    changeCommitted( item );
}

void RssResource::itemRemoved( const Akonadi::Item &item )
{
    kDebug() << "Item id:" << item.id();
    changeCommitted( item );
}

KRss::Feed::Id RssResource::addFeed( const QString &xmlUrl, const QString &subscriptionLabel )
{
    FeedCollection feed;
    feed.setParent( m_rootCollection );
    feed.setContentMimeTypes( QStringList( "application/rss+xml" ) );
    feed.setRemoteId( KRandom::randomString( 10 ) );
    feed.setName( xmlUrl + '_' + feed.remoteId() );
    feed.setXmlUrl( xmlUrl );
    feed.addSubscriptionLabel( subscriptionLabel );

    AddFeedJob *job = new AddFeedJob( feed );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedCreationDone( KJob* ) ) );
    job->start();

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();

    dump();
    return 0;
}

void RssResource::fetchFeed( const KRss::Feed::Id& id )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return;
    }

    if ( m_fetchJobs.contains( id ) ) {
        kWarning() << "Already fetching feed:" << id;
        return;
    }

    FeedCollection feed = m_feeds.value( id );
    kDebug() << "Fetch request for:" << id;
    emit fetchStarted( id );
    FetchFeedJob *job = new FetchFeedJob( feed );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotParallelFeedFetchDone( KJob* ) ) );
    connect( job, SIGNAL( progress( KRss::FeedCollection::Id, unsigned long ) ),
             this, SLOT( slotFetchProgress( KRss::FeedCollection::Id, unsigned long ) ) );
    m_fetchJobs[ id ] = job;
    job->start();
}

void RssResource::abortFetch( const KRss::Feed::Id& id )
{
    if ( !m_fetchJobs.contains( id ) ) {
        kWarning() << "This feed is not being fetched:" << id;
        return;
    }

    FetchFeedJob *job = m_fetchJobs.take( id );
    job->abort();
}

bool RssResource::removeFeed( const KRss::Feed::Id& id )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return false;
    }

    Akonadi::CollectionDeleteJob *job = new Akonadi::CollectionDeleteJob( m_feeds.value( id ) );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedDeletionDone( KJob* ) ) );

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();
    m_feedsInProcess[ job ] = m_feeds.value( id );

    return false;
}

bool RssResource::subscribe( const KRss::Feed::Id& id, const QString &subscriptionLabel )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return false;
    }

    FeedCollection feed = m_feeds.value( id );
    feed.addSubscriptionLabel( subscriptionLabel );
    Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( feed, this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedModificationDone( KJob* ) ) );

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();
    m_feedsInProcess[ job ] = feed;

    return false;
}

bool RssResource::unsubscribe( const KRss::Feed::Id& id, const QString &subscriptionLabel )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return false;
    }

    FeedCollection feed = m_feeds.value( id );
    feed.removeSubscriptionLabel( subscriptionLabel );
    Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( feed, this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedModificationDone( KJob* ) ) );

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();
    m_feedsInProcess[ job ] = feed;

    return false;
}

QStringList RssResource::subscriptions( const KRss::Feed::Id& id )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return QStringList();
    }

    return m_feeds.value( id ).subscriptionLabels();
}

// TODO(dmitry) should use TagId instead of QString
bool RssResource::addTag( const KRss::Feed::Id& id, const QString &tag )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return false;
    }

    FeedCollection feed = m_feeds.value( id );
    feed.addTag( Tag::idFromString( tag ) );
    Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( feed, this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedModificationDone( KJob* ) ) );

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();
    m_feedsInProcess[ job ] = feed;

    return false;
}

// TODO(dmitry) should use TagId instead of QString
bool RssResource::removeTag( const KRss::Feed::Id& id, const QString &tag )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return false;
    }

    FeedCollection feed = m_feeds.value( id );
    feed.removeTag( Tag::idFromString( tag ) );
    Akonadi::CollectionModifyJob *job = new Akonadi::CollectionModifyJob( feed, this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedModificationDone( KJob* ) ) );

    // send the reply later
    setDelayedReply( true );
    m_replies[ job ] = QDBusContext::message().createReply();
    m_feedsInProcess[ job ] = feed;

    return false;
}

QStringList RssResource::tags( const KRss::Feed::Id& id )
{
    if ( !m_feeds.contains( id ) ) {
        kWarning() << "No such feed: id:" << id;
        return QStringList();
    }

    const QList<TagId> feedTags = m_feeds.value( id ).tags();
    QStringList rawTags;
    Q_FOREACH( const TagId &tag, feedTags ) {
        rawTags.append( Tag::idToString( tag ) );
    }
    return rawTags;
}

bool RssResource::importItems( const QString& xmlUrl, const QString& path ) {
    const FeedCollection c = findFeed( xmlUrl );
    ImportItemsJob *job = new ImportItemsJob( c, this );
    job->setSourceFile( path );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotItemImportDone( KJob* ) ) );
    setDelayedReply( true );
    m_replies.insert( job, QDBusContext::message().createReply() );
    job->start();
    return false;
}

bool RssResource::importOpml( const QString &path, const QString &defaultTag )
{
    ImportOpmlJob *job = new ImportOpmlJob( path, m_rootCollection, this );
    job->setDefaultTag( defaultTag );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotImportDone( KJob* ) ) );
    setDelayedReply( true );
    m_replies.insert( job, QDBusContext::message().createReply() );
    job->start();
    return false;
}

void RssResource::exportOpml( const QString &path )
{
    ExportOpmlJob *job = new ExportOpmlJob( path, m_feeds.values(), this );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotExportDone( KJob* ) ) );
    job->start();
}

bool RssResource::retrieveItem( const Akonadi::Item &item, const QSet<QByteArray> &parts )
{
        Q_UNUSED( item )
        Q_UNUSED( parts )

        // what should i do?
        // no way to retrieve a single item from an RSS feed other than
        // refetching the whole feed and looking up the item in the feed
        // (actually, the feed might not contain this item anymore)
        return false;
}

void RssResource::configure( WId windowId )
{
    // note: this WId trick doesn't work
    const QString path = KFileDialog::getOpenFileNameWId( KUrl::fromPath( QDir::homePath() ),
                                                        "*.opml|" + i18n( "OPML Document (*.opml)" ),
                                                        windowId, i18n( "Select an OPML Document" ) );

    importOpml( path, QString() );
}

void RssResource::retrieveCollections()
{
    collectionsRetrievedIncremental( Akonadi::Collection::List(), Akonadi::Collection::List() );
}

void RssResource::retrieveItems( const Akonadi::Collection &col )
{
    FeedCollection feed( col );
    Q_ASSERT( !feed.xmlUrl().isEmpty() );

    kDebug() << "Retrieving items for" << feed.title();
    kDebug() << "Xml Url:" << feed.xmlUrl();

    FetchFeedJob *job = new FetchFeedJob( feed );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotBlockingFeedFetchDone( KJob* ) ) );
    connect( job, SIGNAL( progress( KRss::FeedCollection::Id, unsigned long ) ),
             this, SLOT( slotFetchProgress( KRss::FeedCollection::Id, unsigned long ) ) );
    job->start();
}

void RssResource::slotTagProviderRetrievalDone( KJob *job )
{
    kDebug() << "Tag provider retrieval done";
    if ( job->error() ) {
        kWarning() << job->errorString();
        return;
    }

    Akonadi::CollectionFetchJob *ljob = new Akonadi::CollectionFetchJob( FeedCollection::root(),
                                                                         Akonadi::CollectionFetchJob::Recursive );
    ljob->setResource( identifier() );
    connect( ljob, SIGNAL( result( KJob* ) ), this, SLOT( slotListingDone( KJob* ) ) );
}

void RssResource::slotListingDone( KJob *job )
{
    kDebug() << "Listing done";
    if ( job->error() ) {
        kWarning() << "Collection listing failed";
        kWarning() << job->errorString();
        return;
    }

    QList<Akonadi::Collection> feeds = static_cast<Akonadi::CollectionFetchJob*>( job )->collections();
    Q_FOREACH( const FeedCollection &feed, feeds ) {
        if ( feed.parent() == FeedCollection::root().id() ) {
            m_rootCollection = feed;
        }
        else {
            m_feeds[ feed.id() ] = feed;
        }
    }

    dump();
}

void RssResource::slotFeedCreationDone( KJob *job )
{
    kDebug() << "Feed creation done";
    if ( job->error() ) {
        kWarning() << "Failed to create a new feed";
        kWarning() << job->errorString();
        return;
    }

    FeedCollection feed = static_cast<AddFeedJob*>( job )->feed();
    m_feeds[ feed.id() ] = feed;

    // return the remote id to the caller
    Q_ASSERT( m_replies.contains( job ) );
    QDBusMessage replyMessage = m_replies.take( job );
    replyMessage << feed.id();
    QDBusConnection::sessionBus().send( replyMessage );

    dump();
}

void RssResource::slotFeedModificationDone( KJob *job )
{
    kDebug() << "Feed modification done";

    Q_ASSERT( m_replies.contains( job ) );
    Q_ASSERT( m_feedsInProcess.contains( job ) );
    QDBusMessage replyMessage = m_replies.take( job );
    FeedCollection feed = m_feedsInProcess.take( job );

    if ( job->error() ) {
        kWarning() << "Failed to modify the collection";
        kWarning() << job->errorString();
        replyMessage << false;
    }
    else {
        m_feeds[ feed.id() ] = feed;
        replyMessage << true;
    }

    // return the result to the caller
    QDBusConnection::sessionBus().send( replyMessage );

    dump();
}

void RssResource::slotFeedDeletionDone( KJob *job )
{
    kDebug() << "Feed deletion done";

    Q_ASSERT( m_replies.contains( job ) );
    Q_ASSERT( m_feedsInProcess.contains( job ) );
    QDBusMessage replyMessage = m_replies.take( job );
    FeedCollection feed = m_feedsInProcess.take( job );

    if ( job->error() ) {
        kWarning() << "Failed to modify the collection";
        kWarning() << job->errorString();
        replyMessage << false;
    }
    else {
        m_feeds.remove( feed.id() );
        replyMessage << true;
    }

    // return the result to the caller
    QDBusConnection::sessionBus().send( replyMessage );

    dump();
}

void RssResource::slotBlockingFeedFetchDone( KJob *job )
{
    kDebug() << "Blocking feed fetch done";

    FeedCollection feed = static_cast<FetchFeedJob *>( job )->feed();
    if ( job->error() ) {
        kWarning() << "Failed to sync the feed";
        kWarning() << job->errorString();
        if ( job->error() == FetchFeedJob::Aborted )
            emit fetchAborted( feed.id() );
        else
            emit fetchFailed( feed.id(), job->errorString() );
    }
    else {
        emit fetchFinished( feed.id() );
    }

    m_fetchJobs.remove( feed.id() );

    // notify the scheduler of the finished operation
    itemsRetrievalDone();
}

void RssResource::slotParallelFeedFetchDone( KJob *job )
{
    kDebug() << "Parallel feed fetch done";

    FeedCollection feed = static_cast<FetchFeedJob *>( job )->feed();
    if ( job->error() ) {
        kWarning() << "Failed to sync the feed";
        kWarning() << job->errorString();
        if ( job->error() == FetchFeedJob::Aborted )
            emit fetchAborted( feed.id() );
        else
            emit fetchFailed( feed.id(), job->errorString() );
    }
    else {
        emit fetchFinished( feed.id() );
    }

    m_fetchJobs.remove( feed.id() );
}

void RssResource::slotImportDone( KJob *job )
{
    kDebug() << "Import done";

    Q_ASSERT( m_replies.contains( job ) );
    QDBusMessage replyMessage = m_replies.take( job );
    replyMessage << !job->error();

    if ( job->error() ) {
        kWarning() << "Failed to import the OPML document";
        kWarning() << job->errorString();
    }

    if ( !job->error() && !m_rootCollection.isValid() ) {
        ImportOpmlJob *import = static_cast<ImportOpmlJob *>( job );
        m_rootCollection = import->importedRootCollection();
        m_feeds.unite( import->importedFeeds() );
    }

    QDBusConnection::sessionBus().send( replyMessage );
}


void RssResource::slotItemImportDone( KJob *job )
{

    Q_ASSERT( m_replies.contains( job ) );
    QDBusMessage replyMessage = m_replies.take( job );
    replyMessage << !job->error();

    if ( job->error() ) {
        kWarning() << "Item import failed.";
        kWarning() << job->errorString();
    }

    QDBusConnection::sessionBus().send( replyMessage );
}

void RssResource::slotExportDone( KJob *job )
{
    kDebug() << "Export done";

    if ( job->error() ) {
        kWarning() << "Failed to export the OPML document";
        kWarning() << job->errorString();
    }
}

void RssResource::slotFetchProgress( FeedCollection::Id id, unsigned long percentage )
{
    kDebug() << "Fetch progress for id:" << id << ", percentage:" << percentage;

    emit fetchPercent( id, percentage );
}

FeedCollection RssResource::findFeed( const QString& xmlUrl ) const {
    Q_FOREACH( const FeedCollection& i, m_feeds.values() )
    if ( i.xmlUrl() == xmlUrl )
            return i;
    return FeedCollection();
}

void RssResource::dump()
{
    kDebug() << "Internal data:";
    kDebug() << "Root collection:" << m_rootCollection.name() << ", id:" << m_rootCollection.id();

    kDebug() << "Feeds:";
    QHashIterator<FeedCollection::Id, FeedCollection> it_feed( m_feeds );
    while ( it_feed.hasNext() ) {
        it_feed.next();
        kDebug() << "Feed title:" << it_feed.value().title() << ", id:" << it_feed.key();
    }

    kDebug() << "Feeds in process:";
    QHashIterator<KJob*, FeedCollection> it_process( m_feedsInProcess );
    while ( it_process.hasNext() ) {
        it_process.next();
        kDebug() << "Job:" << it_process.key() << ", feed id:" << it_process.value().id();
    }

    kDebug() << "Fetch jobs:";
    QHashIterator<FeedCollection::Id, FetchFeedJob*> it_fetch( m_fetchJobs );
    while ( it_fetch.hasNext() ) {
        it_fetch.next();
        kDebug() << "Feed id:" << it_fetch.key() << ", fetch job:" << it_fetch.value();
    }

    kDebug() << "Pending D-Bus reply messages:";
    QHashIterator<KJob*, QDBusMessage> it_reply( m_replies );
    while ( it_reply.hasNext() ) {
        it_reply.next();
        kDebug() << "Job:" << it_reply.key() << ", message:" << it_reply.value();
    }
}

AKONADI_RESOURCE_MAIN(RssResource)

#include "rssresource.moc"
