/*
    Copyright (C) 2009    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "newsgatorjobs.h"
#include "soapmessages.h"

#include <KDE/KIO/Job>
#include <KDE/KLocale>
#include <KDE/KDebug>

using namespace KRss;
using namespace Newsgator;

const QString NewsgatorJob::m_apiToken = "E52FEFD1432342698B507EB2DDACF2DF";

void NewsgatorJob::setUserName( const QString& userName )
{
    m_userName = userName;
}

void NewsgatorJob::setPassword( const QString& password )
{
    m_password = password;
}

void NewsgatorJob::setLocation( const Location& location )
{
    m_location = location;
}



NewsgatorTransferJob::NewsgatorTransferJob( QObject *parent )
    : KJob( parent )
{
}

void NewsgatorTransferJob::setRequestData( const QByteArray& requestData )
{
    m_requestData = requestData;
}

void NewsgatorTransferJob::setEndpoint( const QString& endpoint )
{
    m_endpoint = endpoint;
}

void NewsgatorTransferJob::setSoapMessage( SoapMessage* soapMessage )
{
    m_soapMessage = soapMessage;
}

SoapMessage* NewsgatorTransferJob::soapMessage() const
{
    return m_soapMessage;
}

QString NewsgatorTransferJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case NewsgatorTransferJob::CouldNotContactNewsgator:
            result = i18n( "Couldn't contact the Newsgator service.\n%1", errorText() );
            break;
        case NewsgatorTransferJob::CouldNotPassAuthentication:
            result = i18n( "Authentication failed.\n%1", errorText() );
            break;
        case NewsgatorTransferJob::CouldNotParseSoapMessage:
            result = i18n( "Error occured while paring the SOAP message.\n%1", errorText() );
            break;
        case NewsgatorTransferJob::SoapRequestFailed:
            result = i18n( "SOAP request failed.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void NewsgatorTransferJob::start()
{
    Q_ASSERT( m_soapMessage );

    KUrl url( m_endpoint );
    url.setUser( m_userName );
    url.setPass( m_password );
    KIO::TransferJob* job = KIO::http_post( url, m_requestData, KIO::HideProgressInfo );

    job->addMetaData( "UserAgent", "Akonadi Newsgator RSS resource" );
    job->addMetaData( "content-type", "Content-Type: application/soap+xml; charset=utf-8" );

    connect( job, SIGNAL( data( KIO::Job*, const QByteArray& ) ),
             this, SLOT( slotData( KIO::Job*, const QByteArray& ) ) );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotResult( KJob* ) ) );
}

void NewsgatorTransferJob::slotData( KIO::Job *job, const QByteArray& data )
{
    Q_UNUSED( job )
    kDebug() << "Got data:" << data;
    m_soapMessage->addData( data );
}

void NewsgatorTransferJob::slotResult( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotContactNewsgator );
        setErrorText( job->errorString() );
        emitResult();
        return;
    }

    m_soapMessage->parse();
    if ( m_soapMessage->hasError() ) {
        kWarning() << m_soapMessage->errorString();
        setError( CouldNotParseSoapMessage );
        setErrorText( m_soapMessage->errorString() );
        emitResult();
        return;
    }

    if ( m_soapMessage->messageType() == SoapMessage::Fault ) {
        kWarning() << m_soapMessage->soapErrorString();
        setError( SoapRequestFailed );
        setErrorText( m_soapMessage->soapErrorString() );
        emitResult();
        return;
    }

    emitResult();
}

NewsgatorLocationsRetrieveJob::NewsgatorLocationsRetrieveJob( QObject *parent )
    : KJob( parent )
{
}

QList<Location> NewsgatorLocationsRetrieveJob::locations() const
{
    return m_locations;
}

QString NewsgatorLocationsRetrieveJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case NewsgatorLocationsRetrieveJob::CouldNotContactNewsgator:
            result = i18n( "Couldn't contact the Newsgator service.\n%1", errorText() );
            break;
        case NewsgatorLocationsRetrieveJob::CouldNotRetrieveLocations:
            result = i18n( "Couldn't retrieve locations.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void NewsgatorLocationsRetrieveJob::start()
{
    QByteArray requestData;
    SoapRequest request( &requestData );
    request.setApiToken( m_apiToken );
    request.setHeaderNamespace( "http://services.newsgator.com/svc/Location.asmx" );
    request.writeRequestStart();
    request.writeEmptyElement( "http://services.newsgator.com/svc/Location.asmx", "GetLocations" );
    request.writeRequestEnd();

    kDebug() << "Request data:" << requestData;

    NewsgatorTransferJob *job = new NewsgatorTransferJob();
    job->setUserName( m_userName );
    job->setPassword( m_password );
    job->setEndpoint( "http://services.newsgator.com/ngws/svc/Location.asmx" );
    job->setRequestData( requestData );
    job->setSoapMessage( new LocationsResponse() );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotResult( KJob* ) ) );
    job->start();
}

void NewsgatorLocationsRetrieveJob::slotResult( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotRetrieveLocations );
        setErrorText( job->errorString() );
        emitResult();
        return;
    }

    const NewsgatorTransferJob * const njob = qobject_cast<const NewsgatorTransferJob*>( job );
    Q_ASSERT( njob );
    const LocationsResponse * const response = dynamic_cast<const LocationsResponse*>( njob->soapMessage() );
    Q_ASSERT( response );

    m_locations = response->locations();
    emitResult();
}



NewsgatorFeedsRetrieveJob::NewsgatorFeedsRetrieveJob( QObject *parent )
    : FeedsRetrieveJob( parent )
{
}

QList<Akonadi::Collection> NewsgatorFeedsRetrieveJob::feeds() const
{
    return m_feeds;
}

QString NewsgatorFeedsRetrieveJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case NewsgatorFeedsRetrieveJob::CouldNotContactBackend:
            result = i18n( "Couldn't contact the Newsgator service.\n%1", errorText() );
            break;
        case NewsgatorFeedsRetrieveJob::CouldNotRetrieveFeeds:
            result = i18n( "Couldn't retrieve feeds.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void NewsgatorFeedsRetrieveJob::start()
{
    QByteArray requestData;
    SoapRequest request( &requestData );
    request.setApiToken( m_apiToken );
    request.setHeaderNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeRequestStart();
    request.writeStartElement( "GetSubscriptionList" );
    request.writeDefaultNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeTextElement( "location", m_location.name );
    request.writeEmptyElement( "syncToken" );
    request.writeTextElement( "getScores", "false" );
    request.writeEndElement(); // GetSubscriptionList
    request.writeRequestEnd();

    kDebug() << "Request data:" << requestData;

    NewsgatorTransferJob *job = new NewsgatorTransferJob();
    job->setUserName( m_userName );
    job->setPassword( m_password );
    job->setEndpoint( "http://services.newsgator.com/ngws/svc/Subscription.asmx" );
    job->setRequestData( requestData );
    job->setSoapMessage( new SubscriptionsResponse() );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotResult( KJob* ) ) );
    job->start();
}

void NewsgatorFeedsRetrieveJob::slotResult( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotRetrieveFeeds );
        setErrorText( job->errorString() );
        emitResult();
        return;
    }

    const NewsgatorTransferJob * const njob = qobject_cast<const NewsgatorTransferJob*>( job );
    Q_ASSERT( njob );
    const SubscriptionsResponse * const response = dynamic_cast<const SubscriptionsResponse*>( njob->soapMessage() );
    Q_ASSERT( response );

    // create a top-level collection representing the requested NewsGator location
    Akonadi::Collection loc;
    loc.setRemoteId( m_userName + QChar( '-' ) + m_location.name );
    loc.setName( loc.remoteId() );
    loc.setParent( Akonadi::Collection::root() );
    loc.setContentMimeTypes( QStringList( Akonadi::Collection::mimeType() ) );
    m_feeds.append( loc );

    QHashIterator<int, ParsedFeed> f_it( response->feeds() );
    while ( f_it.hasNext() ) {
        f_it.next();
        kDebug() << "Key:" << f_it.key() << ", feed:" << f_it.value().xmlUrl;

        const ParsedFeed parsed = f_it.value();
        FeedCollection feed;

        feed.setTitle( parsed.title );
        feed.setXmlUrl( parsed.attributes.value( "syncXmlUrl" ) );
        feed.setHtmlUrl( parsed.htmlUrl );
        feed.setDescription( parsed.description );
        feed.setFeedType( parsed.type );
        feed.setRemoteId( parsed.attributes.value( "id" ) );
        feed.setName( feed.title() + QChar( '-' ) + feed.remoteId() );
        feed.setParent( loc );
        feed.setContentMimeTypes( QStringList( "application/rss+xml" ) );
        m_feeds.append( feed );
    }

    QHashIterator<QString, QList<int> > t_it( response->tags() );
    while ( t_it.hasNext() ) {
        t_it.next();
        kDebug() << "Tag:" << t_it.key() << ", feeds:" << t_it.value();
    }

    emitResult();
}



NewsgatorFeedCreateJob::NewsgatorFeedCreateJob( QObject *parent )
    : FeedCreateJob( parent ), m_subscriptionId( 0 )
{
}

Akonadi::Collection NewsgatorFeedCreateJob::feed() const
{
    return m_feed;
}

void NewsgatorFeedCreateJob::setXmlUrl( const QString& xmlUrl )
{
    m_xmlUrl = xmlUrl;
}

QString NewsgatorFeedCreateJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case NewsgatorFeedCreateJob::CouldNotContactBackend:
            result = i18n( "Couldn't contact the Newsgator service.\n%1", errorText() );
            break;
        case NewsgatorFeedCreateJob::CouldNotCreateFeed:
            result = i18n( "Couldn't create feed.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void NewsgatorFeedCreateJob::start()
{
    QByteArray requestData;
    SoapRequest request( &requestData );
    request.setApiToken( m_apiToken );
    request.setHeaderNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeRequestStart();
    request.writeStartElement( "AddSubscription" );
    request.writeDefaultNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeTextElement( "XmlUrl", m_xmlUrl );
    request.writeTextElement( "folderId", "0" );
    request.writeEmptyElement( "cred" );
    request.writeEmptyElement( "customTitle" );
    request.writeTextElement( "markRead", "false" );
    request.writeTextElement( "locationId", QString::number( m_location.id ) );
    request.writeEndElement(); // AddSubscription
    request.writeRequestEnd();

    kDebug() << "Request data:" << requestData;

    NewsgatorTransferJob *job = new NewsgatorTransferJob();
    job->setUserName( m_userName );
    job->setPassword( m_password );
    job->setEndpoint( "http://services.newsgator.com/ngws/svc/Subscription.asmx" );
    job->setRequestData( requestData );
    job->setSoapMessage( new SubscriptionsResponse() );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotResult( KJob* ) ) );
    job->start();
}

void NewsgatorFeedCreateJob::slotResult( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotCreateFeed );
        setErrorText( job->errorString() );
        emitResult();
        return;
    }

    const NewsgatorTransferJob * const njob = qobject_cast<const NewsgatorTransferJob*>( job );
    Q_ASSERT( njob );
    const AddSubscriptionResponse * const response = dynamic_cast<const AddSubscriptionResponse*>( njob->soapMessage() );
    Q_ASSERT( response );

    m_subscriptionId = response->subscriptionId();
    kDebug() << "Got subscription id:" << m_subscriptionId;

    NewsgatorFeedsRetrieveJob *rjob = new NewsgatorFeedsRetrieveJob();
    rjob->setUserName( m_userName );
    rjob->setPassword( m_password );
    rjob->setLocation( m_location );
    connect( rjob, SIGNAL( result( KJob* ) ), this, SLOT( slotFeedsRetrieved( KJob* ) ) );
    rjob->start();
}

void NewsgatorFeedCreateJob::slotFeedsRetrieved( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotCreateFeed );
        setErrorText( i18n( "Error occured while retrieving the subscription list: %1",
                      job->errorString() ) );
        emitResult();
        return;
    }

    const NewsgatorFeedsRetrieveJob * const rjob = qobject_cast<const NewsgatorFeedsRetrieveJob*>( job );
    Q_ASSERT( rjob );
    const QList<Akonadi::Collection> cols = rjob->feeds();
    Q_FOREACH( const Akonadi::Collection& col, cols ) {
        if ( col.remoteId() == QString::number( m_subscriptionId ) ) {
            kDebug() << "Found the created subscription:" << col.name();
            kDebug() << "Parent id:" << col.parent();
            kDebug() << "Parent remote id:" << col.parentRemoteId();
            m_feed = col;
            emitResult();
            return;
        }
    }

    setError( CouldNotCreateFeed );
    setErrorText( i18n( "Couldn't find the created feed with id %1 in the subscription list",
                  m_subscriptionId ) );
    kWarning() << errorString();
    emitResult();
}



NewsgatorFeedModifyJob::NewsgatorFeedModifyJob( QObject *parent )
    : FeedModifyJob( parent )
{
}

void NewsgatorFeedModifyJob::setFeed( const Akonadi::Collection& feed )
{
    m_feed = feed;
}

QString NewsgatorFeedModifyJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case NewsgatorFeedModifyJob::CouldNotContactBackend:
            result = i18n( "Couldn't contact the Newsgator service.\n%1", errorText() );
            break;
        case NewsgatorFeedModifyJob::CouldNotModifyFeed:
            result = i18n( "Couldn't modify feed.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void NewsgatorFeedModifyJob::start()
{
    QByteArray requestData;
    SoapRequest request( &requestData );
    request.setApiToken( m_apiToken );
    request.setHeaderNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeRequestStart();
    request.writeStartElement( "RenameSubscription" );
    request.writeDefaultNamespace( "http://services.newsgator.com/svc/Subscription.asmx" );
    request.writeTextElement( "subscriptionId", m_feed.remoteId() );
    request.writeTextElement( "newName", m_feed.title() );
    request.writeTextElement( "locationId", QString::number( m_location.id ) );
    request.writeEndElement(); // RenameSubscription
    request.writeRequestEnd();

    kDebug() << "Request data:" << requestData;

    NewsgatorTransferJob *job = new NewsgatorTransferJob();
    job->setUserName( m_userName );
    job->setPassword( m_password );
    job->setEndpoint( "http://services.newsgator.com/ngws/svc/Subscription.asmx" );
    job->setRequestData( requestData );
    job->setSoapMessage( new SubscriptionsResponse() );
    connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotResult( KJob* ) ) );
    job->start();
}

void NewsgatorFeedModifyJob::slotResult( KJob *job )
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError( CouldNotModifyFeed );
        setErrorText( job->errorString() );
        emitResult();
        return;
    }

    emitResult();
}

#include "newsgatorjobs.moc"
